from gi.repository import Gtk, Gdk, GdkPixbuf

from pychess.ic.FICSObjects import (
    FICSSoughtMatch,
    FICSChallenge,
    get_seek_tooltip_text,
    get_challenge_tooltip_text,
)
from pychess.ic import (
    TYPE_BLITZ,
    TYPE_LIGHTNING,
    TYPE_STANDARD,
    RATING_TYPES,
    TYPE_BULLET,
    TYPE_ONE_MINUTE,
    TYPE_THREE_MINUTE,
    TYPE_FIVE_MINUTE,
    TYPE_FIFTEEN_MINUTE,
    TYPE_FORTYFIVE_MINUTE,
    get_infobarmessage_content,
)
from pychess.perspectives.fics.ParrentListSection import (
    ParrentListSection,
    cmp,
    SEPARATOR,
    ACCEPT,
    ASSESS,
    FOLLOW,
    CHAT,
    CHALLENGE,
    FINGER,
    ARCHIVED,
)
from pychess.Utils.IconLoader import get_pixbuf
from pychess.System import conf, uistuff
from pychess.System.Log import log
from pychess.System.prefix import addDataPrefix
from pychess.widgets import mainwindow
from pychess.widgets.preferencesDialog import SoundTab
from pychess.widgets.InfoBar import InfoBarMessage, InfoBarMessageButton

__title__ = _("Seeks / Challenges")

__icon__ = addDataPrefix("glade/manseek.svg")

__desc__ = _("Handle seeks and challenges")


class Sidepanel(ParrentListSection):
    def load(self, widgets, connection, lounge):
        self.widgets = widgets
        self.connection = connection
        self.lounge = lounge
        self.infobar = lounge.infobar

        __widget__ = lounge.seek_list

        self.messages = {}
        self.seeks = {}
        self.challenges = {}
        self.seekPix = get_pixbuf("glade/seek.png")
        self.chaPix = get_pixbuf("glade/challenge.png")
        self.manSeekPix = get_pixbuf("glade/manseek.png")

        self.widgets["seekExpander"].set_vexpand(False)

        self.tv = self.widgets["seektreeview"]
        self.store = Gtk.ListStore(
            FICSSoughtMatch,
            GdkPixbuf.Pixbuf,
            GdkPixbuf.Pixbuf,
            str,
            int,
            str,
            str,
            str,
            int,
            Gdk.RGBA,
            str,
        )

        self.seek_filter = self.store.filter_new()
        self.seek_filter.set_visible_func(self.seek_filter_func)

        self.filter_toggles = {}
        self.filter_buttons = (
            "standard_toggle1",
            "blitz_toggle1",
            "lightning_toggle1",
            "variant_toggle1",
            "computer_toggle1",
        )
        for widget in self.filter_buttons:
            uistuff.keep(self.widgets[widget], widget)
            self.widgets[widget].connect("toggled", self.on_filter_button_toggled)
            initial = conf.get(widget)
            self.filter_toggles[widget] = initial
            self.widgets[widget].set_active(initial)

        self.model = self.seek_filter.sort_new_with_model()
        self.tv.set_model(self.model)

        self.tv.set_model(self.model)
        self.addColumns(
            self.tv,
            "FICSSoughtMatch",
            "",
            "",
            _("Name"),
            _("Rating"),
            _("Rated"),
            _("Type"),
            _("Clock"),
            "gametime",
            "textcolor",
            "tooltip",
            hide=[0, 8, 9, 10],
            pix=[1, 2],
        )
        self.tv.set_search_column(3)
        self.tv.set_tooltip_column(
            10,
        )
        for i in range(0, 2):
            self.tv.get_model().set_sort_func(i, self.pixCompareFunction, i + 1)
        for i in range(2, 8):
            self.tv.get_model().set_sort_func(i, self.compareFunction, i)
        try:
            self.tv.set_search_position_func(self.lowLeftSearchPosFunc, None)
        except AttributeError:
            # Unknow signal name is raised by gtk < 2.10
            pass
        for num in range(2, 7):
            column = self.tv.get_column(num)
            for cellrenderer in column.get_cells():
                column.add_attribute(cellrenderer, "foreground_rgba", 9)
        self.selection = self.tv.get_selection()
        self.lastSeekSelected = None
        self.selection.set_select_function(self.selectFunction, True)
        self.selection.connect("changed", self.onSelectionChanged)
        self.widgets["clearSeeksButton"].connect("clicked", self.onClearSeeksClicked)
        self.widgets["acceptButton"].connect("clicked", self.on_accept)
        self.widgets["declineButton"].connect("clicked", self.onDeclineClicked)
        self.tv.connect("row-activated", self.row_activated)
        self.tv.connect("button-press-event", self.button_press_event)

        self.connection.seeks.connect("FICSSeekCreated", self.onAddSeek)
        self.connection.seeks.connect("FICSSeekRemoved", self.onRemoveSeek)
        self.connection.challenges.connect("FICSChallengeIssued", self.onChallengeAdd)
        self.connection.challenges.connect(
            "FICSChallengeRemoved", self.onChallengeRemove
        )
        self.connection.glm.connect("our-seeks-removed", self.our_seeks_removed)
        self.connection.glm.connect("assessReceived", self.onAssessReceived)
        self.connection.bm.connect("playGameCreated", self.onPlayingGame)
        self.connection.bm.connect("curGameEnded", self.onCurGameEnded)

        def get_sort_order(modelsort):
            identity, order = modelsort.get_sort_column_id()
            if identity is None or identity < 0:
                identity = 0
            else:
                identity += 1
            if order == Gtk.SortType.DESCENDING:
                identity = -1 * identity
            return identity

        def set_sort_order(modelsort, value):
            if value != 0:
                order = Gtk.SortType.ASCENDING if value > 0 else Gtk.SortType.DESCENDING
                modelsort.set_sort_column_id(abs(value) - 1, order)

        uistuff.keep(
            self.model,
            "seektreeview_sort_order_col",
            get_sort_order,
            lambda modelsort, value: set_sort_order(modelsort, value),
        )

        self.createLocalMenu(
            (ACCEPT, ASSESS, CHALLENGE, CHAT, FOLLOW, SEPARATOR, FINGER, ARCHIVED)
        )
        self.assess_sent = False

        return __widget__

    def seek_filter_func(self, model, iter, data):
        sought_match = model[iter][0]
        is_computer = sought_match.player.isComputer()
        is_standard = (
            sought_match.game_type.rating_type
            in (TYPE_STANDARD, TYPE_FIFTEEN_MINUTE, TYPE_FORTYFIVE_MINUTE)
            and not is_computer
        )
        is_blitz = (
            sought_match.game_type.rating_type
            in (TYPE_BLITZ, TYPE_THREE_MINUTE, TYPE_FIVE_MINUTE)
            and not is_computer
        )
        is_lightning = (
            sought_match.game_type.rating_type
            in (TYPE_LIGHTNING, TYPE_BULLET, TYPE_ONE_MINUTE)
            and not is_computer
        )
        is_variant = (
            sought_match.game_type.rating_type in RATING_TYPES[9:] and not is_computer
        )
        return (
            (self.filter_toggles["computer_toggle1"] and is_computer)
            or (self.filter_toggles["standard_toggle1"] and is_standard)
            or (self.filter_toggles["blitz_toggle1"] and is_blitz)
            or (self.filter_toggles["lightning_toggle1"] and is_lightning)
            or (self.filter_toggles["variant_toggle1"] and is_variant)
        )

    def on_filter_button_toggled(self, widget):
        for button in self.filter_buttons:
            self.filter_toggles[button] = self.widgets[button].get_active()
        self.seek_filter.refilter()

    def onAssessReceived(self, glm, assess):
        if self.assess_sent:
            self.assess_sent = False
            dialog = Gtk.MessageDialog(
                mainwindow(), type=Gtk.MessageType.INFO, buttons=Gtk.ButtonsType.OK
            )
            dialog.set_title(_("Assess"))
            dialog.set_markup(_("Effect on ratings by the possible outcomes"))
            grid = Gtk.Grid()
            grid.set_column_homogeneous(True)
            grid.set_row_spacing(12)
            grid.set_row_spacing(12)
            name0 = Gtk.Label()
            name0.set_markup("<b>%s</b>" % assess["names"][0])
            name1 = Gtk.Label()
            name1.set_markup("<b>%s</b>" % assess["names"][1])
            grid.attach(Gtk.Label(label=""), 0, 0, 1, 1)
            grid.attach(name0, 1, 0, 1, 1)
            grid.attach(name1, 2, 0, 1, 1)
            grid.attach(Gtk.Label(assess["type"]), 0, 1, 1, 1)
            grid.attach(Gtk.Label(assess["oldRD"][0]), 1, 1, 1, 1)
            grid.attach(Gtk.Label(assess["oldRD"][1]), 2, 1, 1, 1)
            grid.attach(Gtk.Label(_("Win:")), 0, 2, 1, 1)
            grid.attach(Gtk.Label(assess["win"][0]), 1, 2, 1, 1)
            grid.attach(Gtk.Label(assess["win"][1]), 2, 2, 1, 1)
            grid.attach(Gtk.Label(_("Draw:")), 0, 3, 1, 1)
            grid.attach(Gtk.Label(assess["draw"][0]), 1, 3, 1, 1)
            grid.attach(Gtk.Label(assess["draw"][1]), 2, 3, 1, 1)
            grid.attach(Gtk.Label(_("Loss:")), 0, 4, 1, 1)
            grid.attach(Gtk.Label(assess["loss"][0]), 1, 4, 1, 1)
            grid.attach(Gtk.Label(assess["loss"][1]), 2, 4, 1, 1)
            grid.attach(Gtk.Label(_("New RD:")), 0, 5, 1, 1)
            grid.attach(Gtk.Label(assess["newRD"][0]), 1, 5, 1, 1)
            grid.attach(Gtk.Label(assess["newRD"][1]), 2, 5, 1, 1)
            grid.show_all()
            dialog.get_message_area().add(grid)
            dialog.run()
            dialog.destroy()

    def getSelectedPlayer(self):
        model, sel_iter = self.tv.get_selection().get_selected()
        if sel_iter is not None:
            sought = model.get_value(sel_iter, 0)
            return sought.player

    def textcolor_normal(self):
        style_ctxt = self.tv.get_style_context()
        return style_ctxt.get_color(Gtk.StateFlags.NORMAL)

    def textcolor_selected(self):
        style_ctxt = self.tv.get_style_context()
        return style_ctxt.get_color(Gtk.StateFlags.INSENSITIVE)

    def selectFunction(self, selection, model, path, is_selected, data):
        if model[path][9] == self.textcolor_selected():
            return False
        else:
            return True

    def __isAChallengeOrOurSeek(self, row):
        sought = row[0]
        textcolor = row[9]
        red0, green0, blue0 = textcolor.red, textcolor.green, textcolor.blue
        selected = self.textcolor_selected()
        red1, green1, blue1 = selected.red, selected.green, selected.blue
        if (isinstance(sought, FICSChallenge)) or (
            red0 == red1 and green0 == green1 and blue0 == blue1
        ):
            return True
        else:
            return False

    def compareFunction(self, model, iter0, iter1, column):
        row0 = list(model[model.get_path(iter0)])
        row1 = list(model[model.get_path(iter1)])
        is_ascending = (
            True
            if self.tv.get_column(column - 1).get_sort_order() is Gtk.SortType.ASCENDING
            else False
        )
        if self.__isAChallengeOrOurSeek(row0) and not self.__isAChallengeOrOurSeek(
            row1
        ):
            if is_ascending:
                return -1
            else:
                return 1
        elif self.__isAChallengeOrOurSeek(row1) and not self.__isAChallengeOrOurSeek(
            row0
        ):
            if is_ascending:
                return 1
            else:
                return -1
        elif column == 7:
            return self.timeCompareFunction(model, iter0, iter1, column)
        else:
            value0 = row0[column]
            value0 = value0.lower() if isinstance(value0, str) else value0
            value1 = row1[column]
            value1 = value1.lower() if isinstance(value1, str) else value1
            return cmp(value0, value1)

    def __updateActiveSeeksLabel(self):
        count = len(self.seeks) + len(self.challenges)
        self.widgets["activeSeeksLabel"].set_text(_("Active seeks: %d") % count)

    def onAddSeek(self, seeks, seek):
        log.debug("%s" % seek, extra={"task": (self.connection.username, "onAddSeek")})
        pix = self.seekPix if seek.automatic else self.manSeekPix
        textcolor = (
            self.textcolor_selected()
            if seek.player.name == self.connection.getUsername()
            else self.textcolor_normal()
        )
        seek_ = [
            seek,
            seek.player.getIcon(gametype=seek.game_type),
            pix,
            seek.player.name + seek.player.display_titles(),
            seek.player_rating,
            seek.display_rated,
            seek.game_type.display_text,
            seek.display_timecontrol,
            seek.sortable_time,
            textcolor,
            get_seek_tooltip_text(seek),
        ]

        if textcolor == self.textcolor_selected():
            txi = self.store.prepend(seek_)
            self.tv.scroll_to_cell(self.store.get_path(txi))
            self.widgets["clearSeeksButton"].set_sensitive(True)
        else:
            txi = self.store.append(seek_)
        self.seeks[hash(seek)] = txi
        self.__updateActiveSeeksLabel()

    def onRemoveSeek(self, seeks, seek):
        log.debug(
            "%s" % seek, extra={"task": (self.connection.username, "onRemoveSeek")}
        )
        try:
            treeiter = self.seeks[hash(seek)]
        except KeyError:
            # We ignore removes we haven't added, as it seems fics sends a
            # lot of removes for games it has never told us about
            return
        if self.store.iter_is_valid(treeiter):
            self.store.remove(treeiter)
        del self.seeks[hash(seek)]
        self.__updateActiveSeeksLabel()

    def onChallengeAdd(self, challenges, challenge):
        log.debug(
            "%s" % challenge,
            extra={"task": (self.connection.username, "onChallengeAdd")},
        )
        SoundTab.playAction("aPlayerChecks")

        # TODO: differentiate between challenges and manual-seek-accepts
        # (wait until seeks are comparable FICSSeek objects to do this)
        # Related: http://code.google.com/p/pychess/issues/detail?id=206
        if challenge.adjourned:
            text = _(
                " would like to resume your adjourned <b>%(time)s</b> "
                + "<b>%(gametype)s</b> game."
            ) % {
                "time": challenge.display_timecontrol,
                "gametype": challenge.game_type.display_text,
            }
        else:
            text = _(
                " challenges you to a <b>%(time)s</b> %(rated)s <b>%(gametype)s</b> game"
            ) % {
                "time": challenge.display_timecontrol,
                "rated": challenge.display_rated.lower(),
                "gametype": challenge.game_type.display_text,
            }
            if challenge.color:
                text += _(" where <b>%(player)s</b> plays <b>%(color)s</b>.") % {
                    "player": challenge.player.name,
                    "color": _("white") if challenge.color == "white" else _("black"),
                }
            else:
                text += "."
        content = get_infobarmessage_content(
            challenge.player, text, gametype=challenge.game_type
        )

        def callback(infobar, response, message):
            if response == Gtk.ResponseType.ACCEPT:
                self.connection.om.acceptIndex(challenge.index)
            elif response == Gtk.ResponseType.NO:
                self.connection.om.declineIndex(challenge.index)
            message.dismiss()
            return False

        message = InfoBarMessage(Gtk.MessageType.QUESTION, content, callback)
        message.add_button(InfoBarMessageButton(_("Accept"), Gtk.ResponseType.ACCEPT))
        message.add_button(InfoBarMessageButton(_("Decline"), Gtk.ResponseType.NO))
        message.add_button(
            InfoBarMessageButton(Gtk.STOCK_CLOSE, Gtk.ResponseType.CANCEL)
        )
        self.messages[hash(challenge)] = message
        self.infobar.push_message(message)

        txi = self.store.prepend(
            [
                challenge,
                challenge.player.getIcon(gametype=challenge.game_type),
                self.chaPix,
                challenge.player.name + challenge.player.display_titles(),
                challenge.player_rating,
                challenge.display_rated,
                challenge.game_type.display_text,
                challenge.display_timecontrol,
                challenge.sortable_time,
                self.textcolor_normal(),
                get_challenge_tooltip_text(challenge),
            ]
        )
        self.challenges[hash(challenge)] = txi
        self.__updateActiveSeeksLabel()
        self.widgets["seektreeview"].scroll_to_cell(self.store.get_path(txi))

    def onChallengeRemove(self, challenges, challenge):
        log.debug(
            "%s" % challenge,
            extra={"task": (self.connection.username, "onChallengeRemove")},
        )
        try:
            txi = self.challenges[hash(challenge)]
        except KeyError:
            pass
        else:
            if self.store.iter_is_valid(txi):
                self.store.remove(txi)
            del self.challenges[hash(challenge)]

        try:
            message = self.messages[hash(challenge)]
        except KeyError:
            pass
        else:
            message.dismiss()
            del self.messages[hash(challenge)]
        self.__updateActiveSeeksLabel()

    def our_seeks_removed(self, glm):
        self.widgets["clearSeeksButton"].set_sensitive(False)

    def onDeclineClicked(self, button):
        model, sel_iter = self.tv.get_selection().get_selected()
        if sel_iter is None:
            return
        sought = model.get_value(sel_iter, 0)
        self.connection.om.declineIndex(sought.index)

        try:
            message = self.messages[hash(sought)]
        except KeyError:
            pass
        else:
            message.dismiss()
            del self.messages[hash(sought)]

    def onClearSeeksClicked(self, button):
        self.connection.client.run_command("unseek")
        self.widgets["clearSeeksButton"].set_sensitive(False)

    def row_activated(self, treeview, path, view_column):
        model, sel_iter = self.tv.get_selection().get_selected()
        if sel_iter is None:
            return
        sought = model.get_value(sel_iter, 0)
        if self.lastSeekSelected is None or sought.index != self.lastSeekSelected.index:
            return
        if path != model.get_path(sel_iter):
            return
        self.on_accept(None)

    def onSelectionChanged(self, selection):
        model, sel_iter = selection.get_selected()
        sought = None
        a_seek_is_selected = False
        selection_is_challenge = False
        if sel_iter is not None:
            a_seek_is_selected = True
            sought = model.get_value(sel_iter, 0)
            if isinstance(sought, FICSChallenge):
                selection_is_challenge = True

            # # select sought owner on players tab to let challenge him using right click menu
            # if sought.player in self.lounge.players_tab.players:
            # # we have to undo the iter conversion that was introduced by the filter and sort model
            # iter0 = self.lounge.players_tab.players[sought.player]["ti"]
            # filtered_model = self.lounge.players_tab.player_filter
            # is_ok, iter1 = filtered_model.convert_child_iter_to_iter(iter0)
            # sorted_model = self.lounge.players_tab.model
            # is_ok, iter2 = sorted_model.convert_child_iter_to_iter(iter1)
            # players_selection = self.lounge.players_tab.tv.get_selection()
            # players_selection.select_iter(iter2)
            # self.lounge.players_tab.tv.scroll_to_cell(sorted_model.get_path(iter2))
            # else:
            # print(sought.player, "not in self.lounge.players_tab.players")

        self.lastSeekSelected = sought
        self.widgets["acceptButton"].set_sensitive(a_seek_is_selected)
        self.widgets["declineButton"].set_sensitive(selection_is_challenge)

    def _clear_messages(self):
        for message in self.messages.values():
            message.dismiss()
        self.messages.clear()

    def onPlayingGame(self, bm, game):
        self._clear_messages()
        self.widgets["seekListContent"].set_sensitive(False)
        self.widgets["clearSeeksButton"].set_sensitive(False)
        self.__updateActiveSeeksLabel()

    def onCurGameEnded(self, bm, game):
        self.widgets["seekListContent"].set_sensitive(True)



import collections
from io import StringIO

from pychess.Utils.GameModel import GameModel
from pychess.Utils.const import (
    WAITING_TO_START,
    BLACKWON,
    WHITEWON,
    DRAW,
    FISCHERRANDOMCHESS,
)
from pychess.Utils.logic import getStatus
from pychess.Variants.fischerandom import FischerandomBoard

from .ChessFile import ChessFile, LoadingError


__label__ = _("Simple Chess Position")
__ending__ = "fen"
__append__ = True


def save(handle, model, position=None, flip=False):
    """Saves game to file in fen format"""
    print(position, model.ply, model.lowply)
    print(model.boards)
    print(
        "%s"
        % model.boards[
            -1 if position is None or len(model.boards) == 1 else position
        ].asFen(),
        file=handle,
    )
    output = handle.getvalue() if isinstance(handle, StringIO) else ""
    handle.close()
    return output


def load(handle):
    return FenFile(handle)


class FenFile(ChessFile):
    def __init__(self, handle):
        ChessFile.__init__(self, handle)
        self.fen_is_string = isinstance(handle, StringIO)
        rec = collections.defaultdict(str)
        line = handle.readline().strip()
        rec["Id"] = 0
        rec["Offset"] = 0
        rec["FEN"] = line

        castling = rec["FEN"].split()[2]
        for letter in castling:
            if letter.upper() in "ABCDEFGH":
                rec["Variant"] = FISCHERRANDOMCHESS
                break

        self.games = [
            rec,
        ]
        self.count = 1

    def loadToModel(self, rec, position, model=None):
        if not model:
            model = GameModel()

        if self.fen_is_string:
            rec = self.games[0]

        if isinstance(rec, dict) and "Variant" in rec:
            model.variant = FischerandomBoard

        fen = self.games[0]["FEN"]
        try:
            board = model.variant(setup=fen)
            model.tags["FEN"] = fen
        except SyntaxError as err:
            board = model.variant()
            raise LoadingError(
                _("The game can't be loaded, because of an error parsing FEN"),
                err.args[0],
            )

        model.boards = [board]
        model.variations = [model.boards]
        model.moves = []
        if model.status == WAITING_TO_START:
            status, reason = getStatus(model.boards[-1])
            if status in (BLACKWON, WHITEWON, DRAW):
                model.status, model.reason = status, reason
        return model



import os
import json
from abc import ABCMeta
from collections import UserDict

from gi.repository import Gdk, Gtk, GObject
from gi.types import GObjectMeta

from pychess.perspectives import Perspective, perspective_manager, panel_name
from pychess.System.prefix import addUserConfigPrefix, addDataPrefix
from pychess.System.Log import log
from pychess.widgets import new_notebook, mainwindow
from pychess.widgets.pydock.PyDockTop import PyDockTop
from pychess.widgets.pydock import WEST, SOUTH, CENTER
from pychess.System.prefix import addUserDataPrefix
from pychess.Savers.olv import OLVFile
from pychess.Savers.pgn import PGNFile
from pychess.System.protoopen import protoopen


class Learn(GObject.GObject, Perspective):
    def __init__(self):
        GObject.GObject.__init__(self)
        Perspective.__init__(self, "learn", _("Learn"))
        self.always_on = True

        self.dockLocation = addUserConfigPrefix("pydock-learn.xml")
        self.first_run = True

    def create_toolbuttons(self):
        def on_exit_clicked(button):
            perspective_manager.disable_perspective("learn")

        self.exit_button = Gtk.ToolButton.new_from_stock(Gtk.STOCK_QUIT)
        self.exit_button.set_tooltip_text(_("Quit Learning"))
        self.exit_button.set_label("exit")
        self.exit_button.connect("clicked", on_exit_clicked)

    def init_layout(self):
        perspective_manager.set_perspective_toolbuttons("learn", (self.exit_button,))

        perspective_widget = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        perspective_manager.set_perspective_widget("learn", perspective_widget)

        self.notebooks = {"home": new_notebook()}
        self.main_notebook = self.notebooks["home"]
        for panel in self.sidePanels:
            self.notebooks[panel_name(panel.__name__)] = new_notebook(
                panel_name(panel.__name__)
            )

        self.dock = PyDockTop("learn", self)
        align = Gtk.Alignment()
        align.show()
        align.add(self.dock)
        self.dock.show()
        perspective_widget.pack_start(align, True, True, 0)

        self.notebooks = {"learnhome": new_notebook()}
        self.main_notebook = self.notebooks["learnhome"]
        for panel in self.sidePanels:
            self.notebooks[panel_name(panel.__name__)] = new_notebook(
                panel_name(panel.__name__)
            )

        self.docks["learnhome"] = (
            Gtk.Label(label="learnhome"),
            self.notebooks["learnhome"],
            None,
        )
        for panel in self.sidePanels:
            self.docks[panel_name(panel.__name__)][1] = self.notebooks[
                panel_name(panel.__name__)
            ]

        self.load_from_xml()

        # Default layout of side panels
        first_time_layout = False
        if not os.path.isfile(self.dockLocation):
            first_time_layout = True
            leaf0 = self.dock.dock(
                self.docks["learnhome"][1],
                CENTER,
                self.docks["learnhome"][0],
                "learnhome",
            )
            leaf0.setDockable(False)

            leaf = leaf0.dock(
                self.docks["PuzzlesPanel"][1],
                WEST,
                self.docks["PuzzlesPanel"][0],
                "PuzzlesPanel",
            )
            leaf.dock(
                self.docks["LessonsPanel"][1],
                SOUTH,
                self.docks["LessonsPanel"][0],
                "LessonsPanel",
            )

            leaf = leaf0.dock(
                self.docks["LecturesPanel"][1],
                SOUTH,
                self.docks["LecturesPanel"][0],
                "LecturesPanel",
            )
            leaf.dock(
                self.docks["EndgamesPanel"][1],
                SOUTH,
                self.docks["EndgamesPanel"][0],
                "EndgamesPanel",
            )

        def unrealize(dock):
            dock.saveToXML(self.dockLocation)
            dock._del()

        self.dock.connect("unrealize", unrealize)

        self.dock.show_all()
        perspective_widget.show_all()

        perspective_manager.set_perspective_menuitems(
            "learn", self.menuitems, default=first_time_layout
        )

        log.debug("Learn.__init__: finished")

    def activate(self):
        if self.first_run:
            self.init_layout()
            self.first_run = False

        learn_home = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        box = Gtk.Box()

        self.tv = Gtk.TreeView()

        color = Gdk.RGBA()
        color.parse("lightblue")

        for i, col in enumerate(
            (_("lichess"), _("wtharvey"), _("yacpdb"), _("lessons"))
        ):
            renderer = Gtk.CellRendererProgress()
            renderer.set_orientation(Gtk.Orientation.VERTICAL)
            renderer.props.height = 100
            renderer.props.inverted = True
            renderer.props.cell_background_rgba = color
            column = Gtk.TreeViewColumn(col, renderer, text=i * 2, value=i * 2 + 1)
            self.tv.append_column(column)

        self.store = Gtk.ListStore(str, int, str, int, str, int, str, int)

        self.update_progress(None, None, None)

        self.tv.set_model(self.store)
        self.tv.get_selection().set_mode(Gtk.SelectionMode.NONE)

        puzzles_solving_progress.connect("progress_updated", self.update_progress)
        lessons_solving_progress.connect("progress_updated", self.update_progress)

        box.pack_start(self.tv, False, False, 6)

        label = Gtk.Label(xpad=6, xalign=0)
        label.set_markup("<b>%s</b>" % _("Progress"))
        learn_home.pack_start(label, False, False, 6)

        learn_home.pack_start(box, False, False, 0)

        reset = Gtk.Button(_("Reset my progress"))
        learn_home.pack_start(reset, False, False, 6)

        def on_reset_clicked(button):
            dialog = Gtk.MessageDialog(
                mainwindow(),
                0,
                Gtk.MessageType.QUESTION,
                Gtk.ButtonsType.OK_CANCEL,
                _("You will lose all your progress data!"),
            )
            response = dialog.run()
            if response == Gtk.ResponseType.OK:
                for filename, progress in lessons_solving_progress.items():
                    lessons_solving_progress[filename] = [0] * len(progress)
                for filename, progress in puzzles_solving_progress.items():
                    puzzles_solving_progress[filename] = [0] * len(progress)
                self.update_progress(None, None, None)
            dialog.destroy()

        reset.connect("clicked", on_reset_clicked)

        learn_home.show_all()

        if not self.first_run:
            self.notebooks["learnhome"].remove_page(-1)
        self.notebooks["learnhome"].append_page(learn_home)

        self.panels = [panel.Sidepanel().load(self) for panel in self.sidePanels]

        for panel, instance in zip(self.sidePanels, self.panels):
            if not self.first_run:
                self.notebooks[panel_name(panel.__name__)].remove_page(-1)
            self.notebooks[panel_name(panel.__name__)].append_page(instance)
            instance.show()

        perspective_manager.activate_perspective("learn")

    def update_progress(self, solving_progress, key, progress):
        self.store.clear()

        # Compute cumulative puzzles solving statistics
        solving_progress = puzzles_solving_progress.read_all()

        stat = [0, 0, 0, 0, 0, 0, 0, 0]
        for filename, progress in solving_progress.items():
            if filename.startswith("lichess"):
                stat[0] += len(progress)
                stat[1] += progress.count(1)
            elif filename.startswith("mate_in"):
                stat[2] += len(progress)
                stat[3] += progress.count(1)
            elif filename.endswith(".olv"):
                stat[4] += len(progress)
                stat[5] += progress.count(1)

        # Compute cumulative lessons solving statistics
        solving_progress = lessons_solving_progress.read_all()

        for filename, progress in solving_progress.items():
            stat[6] += len(progress)
            stat[7] += progress.count(1)

        stats = []
        for i in range(4):
            percent = (
                0
                if not stat[i * 2 + 1]
                else round((stat[i * 2 + 1] * 100.0) / stat[i * 2])
            )
            stats.append("%s%%" % percent)
            stats.append(percent)

        self.store.append(stats)


class GObjectMutableMapping(GObjectMeta, ABCMeta):
    """GObject.GObject and UserDict has different metaclasses
    so we have to create this metaclass to avoid
    TypeError: metaclass conflict: the metaclass of a derived class must be a (non-strict) subclass of the metaclasses of all its bases
    """

    pass


class SolvingProgress(GObject.GObject, UserDict, metaclass=GObjectMutableMapping):
    """Book keeping of puzzle/lesson solving progress
    Each dict key is a .pgn/.olv file name
    Values are list of 0/1 values showing a given file puzzles solved or not
    The dict is automatically synced with corresponding puzzles.json/lessons.json files
    """

    __gsignals__ = {
        "progress_updated": (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (
                str,
                object,
            ),
        ),
    }

    def __init__(self, progress_file):
        GObject.GObject.__init__(self)
        UserDict.__init__(self)
        self.progress_file = addUserDataPrefix(progress_file)

    def get_count(self, filename):
        subdir = "puzzles" if self.progress_file.endswith("puzzles.json") else "lessons"
        if filename.lower().endswith(".pgn"):
            chessfile = PGNFile(protoopen(addDataPrefix(f"learn/{subdir}/{filename}")))
            chessfile.limit = 1000
            chessfile.init_tag_database()
        elif filename.lower().endswith(".olv"):
            chessfile = OLVFile(
                protoopen(
                    addDataPrefix(f"learn/{subdir}/{filename}"),
                    encoding="utf-8",
                )
            )
        chessfile.close()
        count = chessfile.count
        return count

    def __getitem__(self, key):
        if os.path.isfile(self.progress_file):
            with open(self.progress_file) as f:
                self.data = json.load(f)
            if key not in self.data:
                self.__setitem__(key, [0] * self.get_count(key))
        else:
            self.__setitem__(key, [0] * self.get_count(key))

        # print("Solved: %s / %s %s" % (self[key].count(1), len(self[key]), key))

        return self.data[key]

    def __setitem__(self, key, value):
        with open(self.progress_file, "w") as f:
            self.data[key] = value
            json.dump(self.data, f)
        self.emit("progress_updated", key, value)

    def read_all(self):
        if os.path.isfile(self.progress_file):
            with open(self.progress_file) as f:
                self.data = json.load(f)
                return self.data
        else:
            return {}


puzzles_solving_progress = SolvingProgress("puzzles.json")
lessons_solving_progress = SolvingProgress("lessons.json")



from pychess.widgets.TaskerManager import tasker
from pychess.perspectives import Perspective


class Welcome(Perspective):
    def __init__(self):
        Perspective.__init__(self, "welcome", _("Welcome"))

        self.default = True
        self.widget.add(tasker)



from math import e

from pychess.ic.FICSObjects import (
    FICSChallenge,
    get_seek_tooltip_text,
    get_challenge_tooltip_text,
)
from pychess.perspectives.fics.ParrentListSection import ParrentListSection
from pychess.System.Log import log
from pychess.System.prefix import addDataPrefix
from pychess.widgets.SpotGraph import SpotGraph

__title__ = _("Seek Graph")

__icon__ = addDataPrefix("glade/manseek.svg")

__desc__ = _("Handle seeks on graphical way")

YMARKS = (800, 1600, 2400)
# YLOCATION = lambda y: min(y / 3000., 3000)
XMARKS = (5, 15)
# XLOCATION = lambda x: e**(-6.579 / (x + 1))


def YLOCATION(y):
    return min(y / 3000.0, 3000)


def XLOCATION(x):
    return e ** (-6.579 / (x + 1))


# This is used to convert increment time to minutes. With a GAME_LENGTH on
# 40, a game on two minutes and twelve secconds will be placed at the same
# X location as a game on 2+12*40/60 = 10 minutes
GAME_LENGTH = 40


class Sidepanel(ParrentListSection):
    def load(self, widgets, connection, lounge):
        self.widgets = widgets
        self.connection = connection

        __widget__ = lounge.seek_graph

        self.graph = SpotGraph()

        for rating in YMARKS:
            self.graph.addYMark(YLOCATION(rating), str(rating))
        for mins in XMARKS:
            self.graph.addXMark(XLOCATION(mins), str(mins) + _(" min"))

        self.widgets["graphDock"].add(self.graph)
        self.graph.show()
        self.graph.connect("spotClicked", self.onSpotClicked)

        self.connection.seeks.connect("FICSSeekCreated", self.onAddSought)
        self.connection.seeks.connect("FICSSeekRemoved", self.onRemoveSought)
        self.connection.challenges.connect("FICSChallengeIssued", self.onAddSought)
        self.connection.challenges.connect("FICSChallengeRemoved", self.onRemoveSought)
        self.connection.bm.connect("playGameCreated", self.onPlayingGame)
        self.connection.bm.connect("curGameEnded", self.onCurGameEnded)

        return __widget__

    def onSpotClicked(self, graph, name):
        self.connection.bm.play(name)

    def onAddSought(self, manager, sought):
        log.debug(
            "%s" % sought, extra={"task": (self.connection.username, "onAddSought")}
        )
        x_loc = XLOCATION(
            float(sought.minutes) + float(sought.inc) * GAME_LENGTH / 60.0
        )
        y_loc = YLOCATION(float(sought.player_rating))
        if (sought.rated) and ("(C)" in sought.player.long_name()):
            type_ = 2
        elif not (sought.rated) and ("(C)" in sought.player.long_name()):
            type_ = 3
        elif sought.rated:
            type_ = 0
        else:
            type_ = 1

        if isinstance(sought, FICSChallenge):
            tooltip_text = get_challenge_tooltip_text(sought)
        else:
            tooltip_text = get_seek_tooltip_text(sought)
        self.graph.addSpot(sought.index, tooltip_text, x_loc, y_loc, type_)

    def onRemoveSought(self, manager, sought):
        log.debug(
            "%s" % sought, extra={"task": (self.connection.username, "onRemoveSought")}
        )
        self.graph.removeSpot(sought.index)

    def onPlayingGame(self, bm, game):
        self.widgets["seekGraphContent"].set_sensitive(False)

    def onCurGameEnded(self, bm, game):
        self.widgets["seekGraphContent"].set_sensitive(True)



"""The task of this perspective, is to save, load and init new games"""

import os
import subprocess
import tempfile

from collections import defaultdict
from io import StringIO

from gi.repository import Gtk
from gi.repository import GObject

from pychess.Savers.ChessFile import LoadingError
from pychess.Savers import epd, fen, pgn, olv, png, database, html, txt
from pychess.System import conf
from pychess.System.Log import log
from pychess.System.protoopen import isWriteable
from pychess.System.uistuff import GladeWidgets
from pychess.System.prefix import addUserConfigPrefix
from pychess.Savers.pgn import parseDateTag
from pychess.Utils.const import (
    UNFINISHED_STATES,
    ABORTED,
    ABORTED_AGREEMENT,
    LOCAL,
    ARTIFICIAL,
    MENU_ITEMS,
)
from pychess.Utils.Offer import Offer
from pychess.widgets import gamewidget, mainwindow, new_notebook
from pychess.widgets.gamenanny import game_nanny
from pychess.perspectives import Perspective, perspective_manager, panel_name
from pychess.widgets.pydock.PyDockTop import PyDockTop
from pychess.widgets.pydock.__init__ import CENTER, EAST, SOUTH
from pychess.ic.ICGameModel import ICGameModel

enddir = {}
savers = (pgn, epd, fen, olv, png, html, txt)  # chessalpha2 is broken

saveformats = Gtk.ListStore(str, str, GObject.TYPE_PYOBJECT)
exportformats = Gtk.ListStore(str, str, GObject.TYPE_PYOBJECT)

auto = _("Detect type automatically")
saveformats.append([auto, "", None])
exportformats.append([auto, "", None])

for saver in savers:
    label, ending = saver.__label__, saver.__ending__
    endstr = "(%s)" % ending
    enddir[ending] = saver
    if hasattr(saver, "load"):
        saveformats.append([label, endstr, saver])
    else:
        exportformats.append([label, endstr, saver])


class Games(GObject.GObject, Perspective):
    __gsignals__ = {"gmwidg_created": (GObject.SignalFlags.RUN_FIRST, None, (object,))}

    def __init__(self):
        GObject.GObject.__init__(self)
        Perspective.__init__(self, "games", _("Games"))

        self.notebooks = {}
        self.first_run = True
        self.gamewidgets = set()
        self.gmwidg_cids = {}
        self.board_cids = {}
        self.notify_cids = defaultdict(list)

        self.key2gmwidg = {}
        self.key2cid = {}

        self.dock = None
        self.dockAlign = None
        self.dockLocation = addUserConfigPrefix("pydock.xml")

    async def generalStart(self, gamemodel, player0tup, player1tup, loaddata=None):
        """The player tuples are:
        (The type af player in a System.const value,
        A callable creating the player,
        A list of arguments for the callable,
        A preliminary name for the player)

        If loaddata is specified, it should be a tuple of:
        (A text uri or fileobj,
        A Savers.something module with a load function capable of loading it,
        An int of the game in file you want to load,
        The position from where to start the game)
        """

        log.debug(f"Games.generalStart: {gamemodel}\n {player0tup}\n {player1tup}")
        gmwidg = gamewidget.GameWidget(gamemodel, self)
        self.gamewidgets.add(gmwidg)
        self.gmwidg_cids[gmwidg] = gmwidg.connect("game_close_clicked", self.closeGame)

        # worker.publish((gmwidg,gamemodel))
        self.attachGameWidget(gmwidg)
        game_nanny.nurseGame(gmwidg, gamemodel)
        log.debug("Games.generalStart: -> emit gmwidg_created: %s" % (gmwidg))
        self.emit("gmwidg_created", gmwidg)
        log.debug("Games.generalStart: <- emit gmwidg_created: %s" % (gmwidg))

        # Initing players

        def xxxset_name(none, player, key, alt):
            player.setName(conf.get(key, alt))

        players = []
        for i, playertup in enumerate((player0tup, player1tup)):
            type, func, args, prename = playertup
            if type != LOCAL:
                if type == ARTIFICIAL:
                    player = await func(*args)
                else:
                    player = func(*args)
                players.append(player)
                # if type == ARTIFICIAL:
                #    def readyformoves (player, color):
                #        gmwidg.setTabText(gmwidg.display_text))
                #    players[i].connect("readyForMoves", readyformoves, i)
            else:
                # Until PyChess has a proper profiles system, as discussed on the
                # issue tracker, we need to give human players special treatment
                player = func(gmwidg, *args)
                players.append(player)
        assert len(players) == 2
        if player0tup[0] == ARTIFICIAL and player1tup[0] == ARTIFICIAL:

            def emit_action(board, action, player, param, gmwidg):
                if gmwidg.isInFront():
                    gamemodel.curplayer.emit("offer", Offer(action, param=param))

            self.board_cids[gmwidg.board] = gmwidg.board.connect(
                "action", emit_action, gmwidg
            )

        log.debug("Games.generalStart: -> gamemodel.setPlayers(): %s" % (gamemodel))
        gamemodel.setPlayers(players)
        log.debug("Games.generalStart: <- gamemodel.setPlayers(): %s" % (gamemodel))

        # Forward information from the engines
        for playertup, tagname in ((player0tup, "WhiteElo"), (player1tup, "BlackElo")):
            if playertup[0] == ARTIFICIAL:
                elo = playertup[2][0].get("elo")
                if elo:
                    gamemodel.tags[tagname] = elo

        # Starting
        if loaddata:
            try:
                uri, loader, gameno, position = loaddata
                gamemodel.loadAndStart(uri, loader, gameno, position)
                if position != gamemodel.ply and position != -1:
                    gmwidg.board.view.shown = position
            except LoadingError as e:
                d = Gtk.MessageDialog(
                    mainwindow(),
                    type=Gtk.MessageType.WARNING,
                    buttons=Gtk.ButtonsType.OK,
                )
                d.set_markup(_("<big><b>Error loading game</big></b>"))
                d.format_secondary_text(", ".join(str(a) for a in e.args))
                d.show()
                d.destroy()

        else:
            if gamemodel.variant.need_initial_board:
                for player in gamemodel.players:
                    player.setOptionInitialBoard(gamemodel)
            log.debug("Games..generalStart: -> gamemodel.start(): %s" % (gamemodel))
            gamemodel.emit("game_loaded", "")
            gamemodel.start()
            log.debug("Games.generalStart: <- gamemodel.start(): %s" % (gamemodel))

        log.debug(
            "Games.generalStart: returning gmwidg=%s\n gamemodel=%s"
            % (gmwidg, gamemodel)
        )

    ################################################################################
    # Saving                                                                       #
    ################################################################################

    def saveGame(self, game, position=None):
        if not game.isChanged():
            return
        if game.uri and isWriteable(game.uri):
            self.saveGameSimple(game.uri, game, position=position)
        else:
            return self.saveGameAs(game, position=position)

    def saveGameSimple(self, uri, game, position=None):
        ending = os.path.splitext(uri)[1]
        if not ending:
            return
        saver = enddir[ending[1:]]
        game.save(uri, saver, append=False, position=position)

    def saveGamePGN(self, game):
        if conf.get("saveOwnGames") and not game.hasLocalPlayer():
            return True

        filename = conf.get("autoSaveFormat")
        filename = filename.replace("#n1", game.tags["White"])
        filename = filename.replace("#n2", game.tags["Black"])
        year, month, day = parseDateTag(game.tags["Date"])
        year = "" if year is None else str(year)
        month = "" if month is None else str(month)
        day = "" if day is None else str(day)
        filename = filename.replace("#y", "%s" % year)
        filename = filename.replace("#m", "%s" % month)
        filename = filename.replace("#d", "%s" % day)
        pgn_path = conf.get("autoSavePath") + "/" + filename + ".pgn"
        append = True
        try:
            if not os.path.isfile(pgn_path):
                # create new file
                with open(pgn_path, "w"):
                    pass
            base_offset = os.path.getsize(pgn_path)

            # save to .sqlite
            database_path = os.path.splitext(pgn_path)[0] + ".sqlite"
            database.save(database_path, game, base_offset)

            # save to .scout
            from pychess.Savers.pgn import scoutfish_path

            if scoutfish_path is not None:
                pgn_text = pgn.save(StringIO(), game)

                tmp = tempfile.NamedTemporaryFile(
                    mode="w", encoding="utf-8", delete=False
                )
                pgnfile = tmp.name
                with tmp.file as f:
                    f.write(pgn_text)

                # create new .scout from pgnfile we are importing
                args = [scoutfish_path, "make", pgnfile, "%s" % base_offset]
                output = subprocess.check_output(args, stderr=subprocess.STDOUT)

                # append it to our existing one
                if output.decode().find("Processing...done") > 0:
                    old_scout = os.path.splitext(pgn_path)[0] + ".scout"
                    new_scout = os.path.splitext(pgnfile)[0] + ".scout"

                    with open(old_scout, "ab") as file1, open(new_scout, "rb") as file2:
                        file1.write(file2.read())

            # TODO: do we realy want to update .bin ? It can be huge/slow!

            # save to .pgn
            game.save(pgn_path, pgn, append)

            return True
        except OSError:
            return False

    def saveGameAs(self, game, position=None, export=False):
        savedialog, savecombo = get_save_dialog(export)

        # Keep running the dialog until the user has canceled it or made an error
        # free operation
        title = _("Save Game") if not export else _("Export position")
        savedialog.set_title(title)
        while True:
            filename = f"{game.players[0]}-{game.players[1]}"
            savedialog.set_current_name(filename.replace(" ", "_"))

            res = savedialog.run()
            if res != Gtk.ResponseType.ACCEPT:
                break

            uri = savedialog.get_filename()
            ending = os.path.splitext(uri)[1]
            if ending.startswith("."):
                ending = ending[1:]
            append = False

            index = savecombo.get_active()
            if index == 0:
                if ending not in enddir:
                    d = Gtk.MessageDialog(
                        mainwindow(),
                        type=Gtk.MessageType.ERROR,
                        buttons=Gtk.ButtonsType.OK,
                    )
                    folder, file = os.path.split(uri)
                    d.set_markup(_("<big><b>Unknown file type '%s'</b></big>") % ending)
                    d.format_secondary_text(
                        _(
                            "Was unable to save '%(uri)s' as PyChess doesn't know the format '%(ending)s'."
                        )
                        % {"uri": uri, "ending": ending}
                    )
                    d.run()
                    d.destroy()
                    continue
                else:
                    saver = enddir[ending]
            else:
                format = exportformats[index] if export else saveformats[index]
                saver = format[2]
                if ending not in enddir or not saver == enddir[ending]:
                    uri += ".%s" % saver.__ending__

            if os.path.isfile(uri) and not os.access(uri, os.W_OK):
                d = Gtk.MessageDialog(
                    mainwindow(), type=Gtk.MessageType.ERROR, buttons=Gtk.ButtonsType.OK
                )
                d.set_markup(_("<big><b>Unable to save file '%s'</b></big>") % uri)
                d.format_secondary_text(
                    _(
                        "You don't have the necessary rights to save the file.\n\
    Please ensure that you have given the right path and try again."
                    )
                )
                d.run()
                d.destroy()
                continue

            if os.path.isfile(uri):
                d = Gtk.MessageDialog(mainwindow(), type=Gtk.MessageType.QUESTION)
                d.add_buttons(
                    Gtk.STOCK_CANCEL,
                    Gtk.ResponseType.CANCEL,
                    _("_Replace"),
                    Gtk.ResponseType.ACCEPT,
                )
                if saver.__append__:
                    d.add_buttons(Gtk.STOCK_ADD, 1)
                d.set_title(_("File exists"))
                folder, file = os.path.split(uri)
                d.set_markup(
                    _(
                        "<big><b>A file named '%s' already exists. Would you like to replace it?</b></big>"
                    )
                    % file
                )
                d.format_secondary_text(
                    _(
                        "The file already exists in '%s'. If you replace it, its content will be overwritten."
                    )
                    % folder
                )
                replaceRes = d.run()
                d.destroy()

                if replaceRes == 1:
                    append = True
                elif replaceRes == Gtk.ResponseType.CANCEL:
                    continue
            else:
                print(repr(uri))

            try:
                flip = self.cur_gmwidg().board.view.rotation > 0
                game.save(uri, saver, append, position, flip)
            except OSError as e:
                d = Gtk.MessageDialog(mainwindow(), type=Gtk.MessageType.ERROR)
                d.add_buttons(Gtk.STOCK_OK, Gtk.ResponseType.OK)
                d.set_title(_("Could not save the file"))
                d.set_markup(
                    _("<big><b>PyChess was not able to save the game</b></big>")
                )
                d.format_secondary_text(
                    _("The error was: %s") % ", ".join(str(a) for a in e.args)
                )
                d.run()
                d.destroy()
                continue

            break

        savedialog.destroy()
        return res

    ################################################################################
    # Closing                                                                      #
    ################################################################################
    def closeAllGames(self, gamewidgets):
        log.debug("Games.closeAllGames")
        response = None
        changedPairs = [
            (gmwidg, gmwidg.gamemodel)
            for gmwidg in gamewidgets
            if gmwidg.gamemodel.isChanged()
        ]
        if len(changedPairs) == 0:
            response = Gtk.ResponseType.OK

        elif len(changedPairs) == 1:
            response = self.closeGame(changedPairs[0][0])
        else:
            markup = (
                "<big><b>"
                + ngettext(
                    "There is %d game with unsaved moves.",
                    "There are %d games with unsaved moves.",
                    len(changedPairs),
                )
                % len(changedPairs)
                + " "
                + _("Save moves before closing?")
                + "</b></big>"
            )

            for gmwidg, game in changedPairs:
                if not gmwidg.gamemodel.isChanged():
                    response = Gtk.ResponseType.OK
                else:
                    if conf.get("autoSave"):
                        x = self.saveGamePGN(game)
                        if x:
                            response = Gtk.ResponseType.OK
                        else:
                            response = None
                            markup = (
                                "<b><big>"
                                + _(
                                    "Unable to save to configured file. \
                                                    Save the games before closing?"
                                )
                                + "</big></b>"
                            )
                            break

            if response is None:
                widgets = GladeWidgets("saveGamesDialog.glade")
                dialog = widgets["saveGamesDialog"]
                heading = widgets["saveGamesDialogHeading"]
                saveLabel = widgets["saveGamesDialogSaveLabel"]
                treeview = widgets["saveGamesDialogTreeview"]

                heading.set_markup(markup)

                liststore = Gtk.ListStore(bool, str)
                treeview.set_model(liststore)
                renderer = Gtk.CellRendererToggle()
                renderer.props.activatable = True
                treeview.append_column(Gtk.TreeViewColumn("", renderer, active=0))
                treeview.append_column(
                    Gtk.TreeViewColumn("", Gtk.CellRendererText(), text=1)
                )
                for gmwidg, game in changedPairs:
                    liststore.append(
                        (
                            True,
                            "{} {} {}".format(
                                game.players[0], _("vs."), game.players[1]
                            ),
                        )
                    )

                def callback(cell, path):
                    if path:
                        liststore[path][0] = not liststore[path][0]
                    saves = len(tuple(row for row in liststore if row[0]))
                    saveLabel.set_text(
                        ngettext("_Save %d document", "_Save %d documents", saves)
                        % saves
                    )
                    saveLabel.set_use_underline(True)

                renderer.connect("toggled", callback)

                callback(None, None)

                while True:
                    response = dialog.run()
                    if response == Gtk.ResponseType.YES:
                        for i in range(len(liststore) - 1, -1, -1):
                            checked, name = liststore[i]
                            if checked:
                                cgmwidg, cgame = changedPairs[i]
                                if self.saveGame(cgame) == Gtk.ResponseType.ACCEPT:
                                    liststore.remove(liststore.get_iter((i,)))
                                    del changedPairs[i]
                                    if cgame.status in UNFINISHED_STATES:
                                        cgame.end(ABORTED, ABORTED_AGREEMENT)
                                    cgame.terminate()
                                    self.delGameWidget(cgmwidg)
                                else:
                                    break
                        else:
                            break
                    else:
                        break
                dialog.destroy()

        if response not in (Gtk.ResponseType.DELETE_EVENT, Gtk.ResponseType.CANCEL):
            pairs = [(gmwidg, gmwidg.gamemodel) for gmwidg in gamewidgets]
            for gmwidg, game in pairs:
                if game.status in UNFINISHED_STATES:
                    game.end(ABORTED, ABORTED_AGREEMENT)
                game.terminate()
                if gmwidg.notebookKey in self.key2gmwidg:
                    self.delGameWidget(gmwidg)

        return response

    def closeGame(self, gmwidg):
        log.debug("Games.closeGame")
        response = None
        if not gmwidg.gamemodel.isChanged():
            response = Gtk.ResponseType.OK
        else:
            markup = (
                "<b><big>"
                + _("Save the current game before you close it?")
                + "</big></b>"
            )
            if conf.get("autoSave"):
                x = self.saveGamePGN(gmwidg.gamemodel)
                if x:
                    response = Gtk.ResponseType.OK
                else:
                    markup = (
                        "<b><big>"
                        + _(
                            "Unable to save to configured file. \
                                            Save the current game before you close it?"
                        )
                        + "</big></b>"
                    )

            if response is None:
                d = Gtk.MessageDialog(mainwindow(), type=Gtk.MessageType.WARNING)
                d.add_button(_("Close _without Saving"), Gtk.ResponseType.OK)
                d.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
                if gmwidg.gamemodel.uri:
                    d.add_button(Gtk.STOCK_SAVE, Gtk.ResponseType.YES)
                else:
                    d.add_button(Gtk.STOCK_SAVE_AS, Gtk.ResponseType.YES)

                gmwidg.bringToFront()

                d.set_markup(markup)
                d.format_secondary_text(
                    _(
                        "It is not possible later to continue the game,\nif you don't save it."
                    )
                )

                response = d.run()
                d.destroy()

            if response == Gtk.ResponseType.YES:
                # Test if cancel was pressed in the save-file-dialog
                if self.saveGame(gmwidg.gamemodel) != Gtk.ResponseType.ACCEPT:
                    response = Gtk.ResponseType.CANCEL

        if response not in (Gtk.ResponseType.DELETE_EVENT, Gtk.ResponseType.CANCEL):
            if gmwidg.gamemodel.status in UNFINISHED_STATES:
                gmwidg.gamemodel.end(ABORTED, ABORTED_AGREEMENT)

            gmwidg.disconnect(self.gmwidg_cids[gmwidg])
            del self.gmwidg_cids[gmwidg]

            for cid in self.notify_cids[gmwidg]:
                conf.notify_remove(cid)
            del self.notify_cids[gmwidg]

            if gmwidg.board in self.board_cids:
                gmwidg.board.disconnect(self.board_cids[gmwidg.board])
                del self.board_cids[gmwidg.board]

            self.delGameWidget(gmwidg)
            self.gamewidgets.remove(gmwidg)
            gmwidg.gamemodel.terminate()

            db_persp = perspective_manager.get_perspective("database")
            if len(self.gamewidgets) == 0:
                for widget in MENU_ITEMS:
                    if (
                        widget in ("copy_pgn", "copy_fen")
                        and db_persp.preview_panel is not None
                    ):
                        continue
                    gamewidget.getWidgets()[widget].set_property("sensitive", False)

        return response

    def delGameWidget(self, gmwidg):
        """Remove the widget from the GUI after the game has been terminated"""
        log.debug("Games.delGameWidget: starting %s" % repr(gmwidg))
        gmwidg.closed = True
        gmwidg.emit("closed")

        called_from_preferences = False
        window_list = Gtk.Window.list_toplevels()
        widgets = gamewidget.getWidgets()
        for window in window_list:
            if window.is_active() and window == widgets["preferences"]:
                called_from_preferences = True
                break

        pageNum = gmwidg.getPageNumber()
        headbook = self.getheadbook()

        if gmwidg.notebookKey in self.key2gmwidg:
            del self.key2gmwidg[gmwidg.notebookKey]

        if gmwidg.notebookKey in self.key2cid:
            headbook.disconnect(self.key2cid[gmwidg.notebookKey])
            del self.key2cid[gmwidg.notebookKey]

        headbook.remove_page(pageNum)
        for notebook in self.notebooks.values():
            notebook.remove_page(pageNum)

        if headbook.get_n_pages() == 1 and conf.get("hideTabs"):
            self.show_tabs(False)

        if headbook.get_n_pages() == 0:
            if not called_from_preferences:
                # If the last (but not the designGW) gmwidg was closed
                # and we are FICS-ing, present the FICS lounge
                perspective_manager.disable_perspective("games")

        gmwidg._del()

    def init_layout(self):
        perspective_widget = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        perspective_manager.set_perspective_widget("games", perspective_widget)

        self.notebooks = {
            "board": new_notebook("board"),
            "buttons": new_notebook("buttons"),
            "messageArea": new_notebook("messageArea"),
        }
        self.main_notebook = self.notebooks["board"]
        for panel in self.sidePanels:
            self.notebooks[panel_name(panel.__name__)] = new_notebook(
                panel_name(panel.__name__)
            )

        # Initing headbook

        align = gamewidget.createAlignment(4, 4, 0, 4)
        align.set_property("yscale", 0)

        headbook = Gtk.Notebook()
        headbook.set_name("headbook")
        headbook.set_scrollable(True)
        align.add(headbook)
        perspective_widget.pack_start(align, False, True, 0)
        self.show_tabs(not conf.get("hideTabs"))

        # Initing center

        centerVBox = Gtk.VBox()

        # The dock

        self.dock = PyDockTop("main", self)
        self.dockAlign = gamewidget.createAlignment(4, 4, 0, 4)
        self.dockAlign.add(self.dock)
        centerVBox.pack_start(self.dockAlign, True, True, 0)
        self.dockAlign.show()
        self.dock.show()

        self.docks["board"] = (Gtk.Label(label="Board"), self.notebooks["board"], None)
        for panel in self.sidePanels:
            self.docks[panel_name(panel.__name__)][1] = self.notebooks[
                panel_name(panel.__name__)
            ]

        self.load_from_xml()

        # Default layout of side panels
        first_time_layout = False
        if not os.path.isfile(self.dockLocation):
            first_time_layout = True
            leaf = self.dock.dock(
                self.docks["board"][1],
                CENTER,
                Gtk.Label(label=self.docks["board"][0]),
                "board",
            )
            self.docks["board"][1].show_all()
            leaf.setDockable(False)

            # S
            epanel = leaf.dock(
                self.docks["bookPanel"][1],
                SOUTH,
                self.docks["bookPanel"][0],
                "bookPanel",
            )
            epanel.default_item_height = 45
            epanel = epanel.dock(
                self.docks["engineOutputPanel"][1],
                CENTER,
                self.docks["engineOutputPanel"][0],
                "engineOutputPanel",
            )

            # NE
            leaf = leaf.dock(
                self.docks["annotationPanel"][1],
                EAST,
                self.docks["annotationPanel"][0],
                "annotationPanel",
            )
            leaf = leaf.dock(
                self.docks["historyPanel"][1],
                CENTER,
                self.docks["historyPanel"][0],
                "historyPanel",
            )
            leaf = leaf.dock(
                self.docks["scorePanel"][1],
                CENTER,
                self.docks["scorePanel"][0],
                "scorePanel",
            )

            # SE
            leaf = leaf.dock(
                self.docks["chatPanel"][1],
                SOUTH,
                self.docks["chatPanel"][0],
                "chatPanel",
            )
            leaf = leaf.dock(
                self.docks["commentPanel"][1],
                CENTER,
                self.docks["commentPanel"][0],
                "commentPanel",
            )

        def unrealize(dock, notebooks):
            # unhide the panel before saving so its configuration is saved correctly
            self.notebooks["board"].get_parent().get_parent().zoomDown()
            dock.saveToXML(self.dockLocation)
            dock._del()

        self.dock.connect("unrealize", unrealize, self.notebooks)

        hbox = Gtk.HBox()

        # Buttons
        self.notebooks["buttons"].set_border_width(4)
        hbox.pack_start(self.notebooks["buttons"], False, True, 0)

        # The message area
        # TODO: If you try to fix this first read issue #958 and 1018
        align = gamewidget.createAlignment(0, 0, 0, 0)
        # sw = Gtk.ScrolledWindow()
        # port = Gtk.Viewport()
        # port.add(self.notebooks["messageArea"])
        # sw.add(port)
        # align.add(sw)
        align.add(self.notebooks["messageArea"])
        hbox.pack_start(align, True, True, 0)

        def ma_switch_page(notebook, gpointer, page_num):
            notebook.props.visible = (
                notebook.get_nth_page(page_num).get_child().props.visible
            )

        self.notebooks["messageArea"].connect("switch-page", ma_switch_page)
        centerVBox.pack_start(hbox, False, True, 0)

        perspective_widget.pack_start(centerVBox, True, True, 0)
        centerVBox.show_all()
        perspective_widget.show_all()

        perspective_manager.set_perspective_menuitems(
            "games", self.menuitems, default=first_time_layout
        )

        conf.notify_add("hideTabs", self.tabsCallback)

        # Connecting headbook to other notebooks

        def hb_switch_page(notebook, gpointer, page_num):
            for notebook in self.notebooks.values():
                notebook.set_current_page(page_num)

            gmwidg = self.key2gmwidg[self.getheadbook().get_nth_page(page_num)]
            if isinstance(gmwidg.gamemodel, ICGameModel):
                primary = "primary " + str(gmwidg.gamemodel.ficsgame.gameno)
                gmwidg.gamemodel.connection.client.run_command(primary)

        headbook.connect("switch-page", hb_switch_page)

        if hasattr(headbook, "set_tab_reorderable"):

            def page_reordered(widget, child, new_num, headbook):
                old_num = self.notebooks["board"].page_num(
                    self.key2gmwidg[child].boardvbox
                )
                if old_num == -1:
                    log.error("Games and labels are out of sync!")
                else:
                    for notebook in self.notebooks.values():
                        notebook.reorder_child(notebook.get_nth_page(old_num), new_num)

            headbook.connect("page-reordered", page_reordered, headbook)

    def adjust_divider(self, diff):
        """Try to move paned (containing board) divider to show/hide captured pieces"""
        if self.dock is None:
            return
        child = self.dock.get_children()[0]
        c1 = child.paned.get_child1()
        if hasattr(c1, "paned"):
            c1.paned.set_position(c1.paned.get_position() + diff)
        else:
            child.paned.set_position(child.paned.get_position() + diff)

    def getheadbook(self):
        if len(self.key2gmwidg) == 0:
            return None
        headbook = self.widget.get_children()[0].get_children()[0].get_child()
        # to help StoryText create widget description
        # headbook.get_tab_label_text = customGetTabLabelText
        return headbook

    def cur_gmwidg(self):
        if len(self.key2gmwidg) == 0:
            return None
        headbook = self.getheadbook()
        notebookKey = headbook.get_nth_page(headbook.get_current_page())
        return self.key2gmwidg[notebookKey]

    def customGetTabLabelText(self, child):
        gmwidg = self.key2gmwidg[child]
        return gmwidg.display_text

    def zoomToBoard(self, view_zoomed):
        if not self.notebooks["board"].get_parent():
            return
        if view_zoomed:
            self.notebooks["board"].get_parent().get_parent().zoomUp()
        else:
            self.notebooks["board"].get_parent().get_parent().zoomDown()

    def show_tabs(self, show):
        head = self.getheadbook()
        if head is None:
            return
        head.set_show_tabs(show)

    def tabsCallback(self, widget):
        head = self.getheadbook()
        if not head:
            return
        if head.get_n_pages() == 1:
            self.show_tabs(not conf.get("hideTabs"))

    def attachGameWidget(self, gmwidg):
        log.debug("attachGameWidget: %s" % gmwidg)
        if self.first_run:
            self.init_layout()
            self.first_run = False
        perspective_manager.activate_perspective("games")

        gmwidg.panels = [panel.Sidepanel().load(gmwidg) for panel in self.sidePanels]
        self.key2gmwidg[gmwidg.notebookKey] = gmwidg
        headbook = self.getheadbook()

        headbook.append_page(gmwidg.notebookKey, gmwidg.tabcontent)
        gmwidg.notebookKey.show_all()

        if hasattr(headbook, "set_tab_reorderable"):
            headbook.set_tab_reorderable(gmwidg.notebookKey, True)

        def callback(notebook, gpointer, page_num, gmwidg):
            if notebook.get_nth_page(page_num) == gmwidg.notebookKey:
                gmwidg.infront()
                if gmwidg.gamemodel.players and gmwidg.gamemodel.isObservationGame():
                    gmwidg.light_on_off(False)
                    text = gmwidg.game_info_label.get_text()
                    gmwidg.game_info_label.set_markup(
                        '<span color="black" weight="bold">%s</span>' % text
                    )

        self.key2cid[gmwidg.notebookKey] = headbook.connect_after(
            "switch-page", callback, gmwidg
        )
        gmwidg.infront()

        align = gamewidget.createAlignment(0, 0, 0, 0)
        align.show()
        align.add(gmwidg.infobar)
        self.notebooks["messageArea"].append_page(align, None)
        self.notebooks["board"].append_page(gmwidg.boardvbox, None)
        gmwidg.boardvbox.show_all()
        for panel, instance in zip(self.sidePanels, gmwidg.panels):
            self.notebooks[panel_name(panel.__name__)].append_page(instance, None)
            instance.show_all()
        self.notebooks["buttons"].append_page(gmwidg.stat_hbox, None)
        gmwidg.stat_hbox.show_all()

        if headbook.get_n_pages() == 1:
            self.show_tabs(not conf.get("hideTabs"))
        else:
            # We should always show tabs if more than one exists
            self.show_tabs(True)

        headbook.set_current_page(-1)

        widgets = gamewidget.getWidgets()
        if headbook.get_n_pages() == 1 and not widgets["show_sidepanels"].get_active():
            self.zoomToBoard(True)


def get_save_dialog(export=False):
    savedialog = Gtk.FileChooserDialog(
        "",
        mainwindow(),
        Gtk.FileChooserAction.SAVE,
        (
            Gtk.STOCK_CANCEL,
            Gtk.ResponseType.CANCEL,
            Gtk.STOCK_SAVE,
            Gtk.ResponseType.ACCEPT,
        ),
    )
    savedialog.set_current_folder(os.path.expanduser("~"))

    # Add widgets to the savedialog
    savecombo = Gtk.ComboBox()
    savecombo.set_name("savecombo")

    crt = Gtk.CellRendererText()
    savecombo.pack_start(crt, True)
    savecombo.add_attribute(crt, "text", 0)

    crt = Gtk.CellRendererText()
    savecombo.pack_start(crt, False)
    savecombo.add_attribute(crt, "text", 1)

    if export:
        savecombo.set_model(exportformats)
    else:
        savecombo.set_model(saveformats)

    savecombo.set_active(1)  # pgn
    savedialog.set_extra_widget(savecombo)

    return savedialog, savecombo


def get_open_dialog():
    opendialog = Gtk.FileChooserDialog(
        _("Open chess file"),
        mainwindow(),
        Gtk.FileChooserAction.OPEN,
        (
            Gtk.STOCK_CANCEL,
            Gtk.ResponseType.CANCEL,
            Gtk.STOCK_OPEN,
            Gtk.ResponseType.OK,
        ),
    )
    opendialog.set_show_hidden(True)
    opendialog.set_select_multiple(True)

    # All chess files filter
    all_filter = Gtk.FileFilter()
    all_filter.set_name(_("All Chess Files"))
    opendialog.add_filter(all_filter)
    opendialog.set_filter(all_filter)

    # Specific filters and save formats
    for ending, saver in enddir.items():
        label = saver.__label__
        endstr = "(%s)" % ending
        f = Gtk.FileFilter()
        f.set_name(label + " " + endstr)
        if hasattr(saver, "load"):
            f.add_pattern("*." + ending)
            all_filter.add_pattern("*." + ending)
            opendialog.add_filter(f)

    return opendialog



from gi.repository import Gtk, WebKit


def open_link(label, url):
    """Gtk.Label() can use this like label.connect("activate-link", open_link)"""
    WebKitBrowser(url)
    return True


class WebKitBrowser:
    def __init__(self, url):
        from pychess.System.uistuff import keepWindowSize

        self.window = Gtk.Window()
        keepWindowSize("webkitbrowser", self.window, (800, 600))

        self.vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.window.add(self.vbox)

        self.box = Gtk.Box()

        self.toolbar = Gtk.Toolbar()
        self.box.pack_start(self.toolbar, False, False, 0)

        self.go_back_button = Gtk.ToolButton(stock_id=Gtk.STOCK_GO_BACK)
        self.toolbar.insert(self.go_back_button, -1)

        self.go_forward_button = Gtk.ToolButton(stock_id=Gtk.STOCK_GO_FORWARD)
        self.toolbar.insert(self.go_forward_button, -1)

        self.go_refresh_button = Gtk.ToolButton(stock_id=Gtk.STOCK_REFRESH)
        self.toolbar.insert(self.go_refresh_button, -1)

        self.url = Gtk.Entry()
        self.box.pack_start(self.url, True, True, 0)

        self.search_entry = Gtk.SearchEntry()
        self.box.pack_start(self.search_entry, False, False, 0)

        self.vbox.pack_start(self.box, False, False, 0)

        self.view = WebKit.WebView()
        self.scrolled_window = Gtk.ScrolledWindow()
        self.scrolled_window.add(self.view)

        self.vbox.pack_start(self.scrolled_window, True, True, 0)

        self.window.show_all()

        self.view.connect("load-committed", self.check_buttons)
        self.view.connect("title-changed", self.change_title)

        self.url.connect("activate", self.go)
        self.search_entry.connect("activate", self.search)
        self.go_back_button.connect("clicked", self.go_back)
        self.go_forward_button.connect("clicked", self.go_forward)
        self.go_refresh_button.connect("clicked", self.refresh)

        self.view.open(url)
        self.view.show()

    def go(self, widget):
        link = self.url.get_text()
        if link.startswith("http://"):
            self.view.open(link)
        else:
            self.view.open("http://" + link)
        self.view.show()

    def search(self, widget):
        text = self.search_entry.get_text()
        text = text.replace(" ", "+")
        self.url.set_text("http://www.google.com/search?q=" + text)
        self.search_entry.set_text("")
        self.go(self)

    def check_buttons(self, widget, data):
        uri = widget.get_main_frame().get_uri()
        self.url.set_text(uri)
        self.go_back_button.set_sensitive(self.view.can_go_back())
        self.go_forward_button.set_sensitive(self.view.can_go_forward())

    def change_title(self, widget, data, arg):
        title = widget.get_main_frame().get_title()
        self.window.set_title(title)

    def go_back(self, widget):
        self.view.go_back()

    def go_forward(self, widget):
        self.view.go_forward()

    def refresh(self, widget):
        self.view.reload()



from time import strftime

from gi.repository import GLib, Gtk, Gdk, GObject, Pango

from pychess.System import uistuff
from pychess.System.prefix import addDataPrefix
from pychess.widgets import insert_formatted
from pychess.ic import FICS_COMMANDS, FICS_HELP


__title__ = _("Console")

__icon__ = addDataPrefix("glade/panel_terminal.svg")

__desc__ = _("Command line interface to the chess server")


class Sidepanel:
    def load(self, widgets, connection, lounge):
        self.connection = connection
        self.consoleView = ConsoleView(self.connection)
        self.consoleView.show_all()
        __widget__ = self.consoleView

        connection.com.connect("consoleMessage", self.onConsoleMessage)

        return __widget__

    @staticmethod
    def filter_unprintable(s):
        return "".join([c for c in s if ord(c) > 31 or ord(c) == 9 or c == "\n"])

    def scroll_to_bottom(self):
        tb_iter = self.consoleView.textbuffer.get_end_iter()
        self.consoleView.readView.scroll_to_iter(tb_iter, 0.00, True, 0.00, 1.00)

    def onConsoleMessage(self, com, lines, ini_lines=None):
        need_scroll = False
        if ini_lines is not None:
            for line in ini_lines:
                self.consoleView.addMessage(line, False)
                need_scroll = True

        for line in lines:
            line = self.filter_unprintable(line.line)
            if (
                line
                and (not line.startswith("<"))
                and (not line.startswith("{Game"))
                and (not line.endswith("available for matches."))
                and line[-12:-5] != "), Bug("
            ):
                self.consoleView.addMessage(line, False)
                need_scroll = True

        if need_scroll:
            # scroll to the bottom but only if we are not scrolled up to read back
            adj = self.consoleView.sw.get_vadjustment()
            if adj.get_value() >= adj.get_upper() - adj.get_page_size() - 1e-12:
                GLib.idle_add(self.scroll_to_bottom)


TYPE_COMMAND, TYPE_HELP, TYPE_USER = 0, 1, 2


class ConsoleView(Gtk.Box):
    __gsignals__ = {
        "messageAdded": (GObject.SignalFlags.RUN_FIRST, None, (str, str, object)),
        "messageTyped": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    }

    def __init__(self, connection):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL, spacing=6)
        self.connection = connection
        self.connection.players.connect("FICSPlayerEntered", self.on_player_entered)
        self.connection.players.connect("FICSPlayerExited", self.on_player_exited)

        # Inits the read view
        self.readView = Gtk.TextView()
        fontdesc = Pango.FontDescription("Monospace 10")
        self.readView.modify_font(fontdesc)

        self.textbuffer = self.readView.get_buffer()
        self.textbuffer.create_tag("text")
        self.textbuffer.create_tag("mytext", weight=Pango.Weight.BOLD)

        self.sw = Gtk.ScrolledWindow()
        self.sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.sw.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
        uistuff.keepDown(self.sw)
        self.sw.add(self.readView)
        self.readView.set_editable(False)
        self.readView.set_cursor_visible(False)
        self.readView.props.wrap_mode = Gtk.WrapMode.WORD
        self.pack_start(self.sw, True, True, 0)

        # Inits entry
        self.history = []
        self.pos = 0

        self.liststore = Gtk.ListStore(str, int)
        for command in FICS_COMMANDS:
            self.liststore.append([command, TYPE_COMMAND])
        for command in FICS_HELP:
            self.liststore.append([command, TYPE_HELP])

        completion = Gtk.EntryCompletion()
        completion.set_model(self.liststore)
        completion.set_text_column(0)

        completion.set_minimum_key_length(2)
        completion.set_popup_set_width(False)

        def match(completion, entrystr, iter, data):
            modelstr = completion.get_model()[iter][0].lower()
            modeltype = completion.get_model()[iter][1]
            parts = entrystr.split()
            if len(parts) == 1 and modeltype == TYPE_COMMAND:
                return modelstr.startswith(entrystr)
            elif len(parts) == 2:
                if parts[0] == "help":
                    return modelstr.startswith(parts[1]) and modeltype == TYPE_HELP
                else:
                    return (
                        parts[0] in FICS_COMMANDS
                        and modelstr.startswith(parts[1].lower())
                        and modeltype == TYPE_USER
                    )

        completion.set_match_func(match, None)

        def on_match_selected(completion, treemodel, treeiter):
            modelstr = treemodel[treeiter][0]
            modeltype = treemodel[treeiter][1]
            entry = completion.get_entry()
            parts = entry.get_text().split()
            if len(parts) == 1 and modeltype == TYPE_COMMAND:
                entry.set_text(modelstr)
                entry.set_position(-1)
                return True
            elif len(parts) == 2:
                entry.set_text(f"{parts[0]} {modelstr}")
                entry.set_position(-1)
                return True

        completion.connect("match-selected", on_match_selected)

        self.entry = Gtk.Entry()
        self.entry.set_completion(completion)

        self.pack_start(self.entry, False, True, 0)

        self.entry.connect("key-press-event", self.onKeyPress)

    def on_player_entered(self, players, new_players):
        for player in new_players:
            self.liststore.append([player.name, TYPE_USER])

    def on_player_exited(self, players, player):
        for row in self.liststore:
            if row[0] == player.name:
                self.liststore.remove(row.iter)
                break

    def addMessage(self, text, my):
        tag = "mytext" if my else "text"
        text_buffer = self.readView.get_buffer()
        tb_iter = text_buffer.get_end_iter()
        # Messages have linebreak before the text. This is opposite to log
        # messages
        if text_buffer.props.text:
            text_buffer.insert(tb_iter, "\n")
        time = strftime("%H:%M:%S")
        text_buffer.insert_with_tags_by_name(tb_iter, "(%s) " % time, tag)
        insert_formatted(self.readView, tb_iter, text, tag=tag)

    def onKeyPress(self, widget, event):
        if event.keyval in list(map(Gdk.keyval_from_name, ("Return", "KP_Enter"))):
            if not event.get_state() & Gdk.ModifierType.CONTROL_MASK:
                buffer = self.entry.get_buffer()
                if buffer.props.text.startswith("pas"):
                    # don't log password changes
                    self.connection.client.telnet.sensitive = True
                self.connection.client.run_command(buffer.props.text, show_reply=True)
                self.emit("messageTyped", buffer.props.text)
                self.addMessage(buffer.props.text, True)

                self.history.append(buffer.props.text)
                buffer.props.text = ""
                self.pos = len(self.history)
                return True

        elif event.keyval == Gdk.keyval_from_name("Up"):
            if self.pos > 0:
                buffer = self.entry.get_buffer()
                self.pos -= 1
                buffer.props.text = self.history[self.pos]
            widget.grab_focus()
            return True

        elif event.keyval == Gdk.keyval_from_name("Down"):
            buffer = self.entry.get_buffer()
            if self.pos == len(self.history) - 1:
                self.pos += 1
                buffer.props.text = ""
            elif self.pos < len(self.history):
                self.pos += 1
                buffer.props.text = self.history[self.pos]
            widget.grab_focus()
            return True



from gi.repository import GObject


class AutoLogOutManager(GObject.GObject):
    __gsignals__ = {"logOut": (GObject.SignalFlags.RUN_FIRST, None, ())}

    def __init__(self, connection):
        GObject.GObject.__init__(self)
        self.connection = connection

        self.connection.expect_line(
            self.onLogOut,
            r"\*\*\*\* Auto-logout because you were idle more than \d+ minutes\. \*\*\*\*",
        )
        self.connection.expect_line(self.onLogOut, r"Logging you out\.")
        self.connection.expect_line(
            self.onLogOut,
            r"\*\*\*\* .+? has arrived - you can't both be logged in\. \*\*\*\*",
        )

    def onLogOut(self, match):
        self.emit("logOut")



import collections
import os
import re
import subprocess
import zipfile

from gi.repository import GLib

from sqlalchemy import select, func
from sqlalchemy.exc import SQLAlchemyError

from pychess.Utils.const import NORMALCHESS, RUNNING, DRAW, WHITEWON, BLACKWON
from pychess.Variants import name2variant
from pychess.System.Log import log
from pychess.System import download_file
from pychess.System.protoopen import protoopen, protosave, PGN_ENCODING
from pychess.Database.model import (
    event,
    site,
    player,
    game,
    annotator,
    tag_game,
    source,
)

# from pychess.System import profile_me

# Editable (on game info dialog) tags
dedicated_tags = (
    "Event",
    "Site",
    "Date",
    "Round",
    "White",
    "Black",
    "WhiteElo",
    "BlackElo",
)

# Other tags stored in game table
other_game_tags = (
    "Result",
    "SetUp",
    "FEN",
    "ECO",
    "Variant",
    "PlyCount",
    "Annotator",
    "offset",
    "offset8",
)

TAG_REGEX = re.compile(r"\[([a-zA-Z0-9_]+)\s+\"(.*)\"\]")

GAME, EVENT, SITE, PLAYER, ANNOTATOR, SOURCE, STAT = range(7)

removeDic = {
    ord("'"): None,
    ord(","): None,
    ord("."): None,
    ord("-"): None,
    ord(" "): None,
}

pgn2Const = {
    "*": RUNNING,
    "?": RUNNING,
    "1/2-1/2": DRAW,
    "1/2": DRAW,
    "1-0": WHITEWON,
    "0-1": BLACKWON,
}


class PgnImport:
    def __init__(self, chessfile, append_pgn=False):
        self.chessfile = chessfile
        self.append_pgn = append_pgn
        self.cancel = False

    def initialize(self):
        self.db_handle = self.chessfile.handle
        self.engine = self.chessfile.engine
        self.conn = self.engine.connect()
        self.CHUNK = 1000

        self.ins_event = event.insert()
        self.ins_site = site.insert()
        self.ins_player = player.insert()
        self.ins_annotator = annotator.insert()
        self.ins_source = source.insert()
        self.ins_game = game.insert()
        self.ins_tag_game = tag_game.insert()

        self.event_dict = {}
        self.site_dict = {}
        self.player_dict = {}
        self.annotator_dict = {}
        self.source_dict = {}

        self.next_id = [0, 0, 0, 0, 0, 0]

        self.next_id[GAME] = self.ini_names(game, GAME)
        self.next_id[EVENT] = self.ini_names(event, EVENT)
        self.next_id[SITE] = self.ini_names(site, SITE)
        self.next_id[PLAYER] = self.ini_names(player, PLAYER)
        self.next_id[ANNOTATOR] = self.ini_names(annotator, ANNOTATOR)
        self.next_id[SOURCE] = self.ini_names(source, SOURCE)

    def get_id(self, name, name_table, field, info=None):
        if not name:
            return None

        orig_name = name
        if field == EVENT:
            name_dict = self.event_dict
            name_data = self.event_data
        elif field == SITE:
            name_dict = self.site_dict
            name_data = self.site_data
        elif field == ANNOTATOR:
            name_dict = self.annotator_dict
            name_data = self.annotator_data
        elif field == SOURCE:
            name_dict = self.source_dict
            name_data = self.source_data
        elif field == PLAYER:
            name_dict = self.player_dict
            name_data = self.player_data
            name = name.title().translate(removeDic)

        if name in name_dict:
            return name_dict[name]

        if field == SOURCE:
            name_data.append({"name": orig_name, "info": info})
        else:
            name_data.append({"name": orig_name})
        name_dict[name] = self.next_id[field]
        self.next_id[field] += 1
        return name_dict[name]

    def ini_names(self, name_table, field):
        if field != GAME and field != STAT:
            name_dict = {
                n.name.title().translate(removeDic): n.id
                for n in self.conn.execute(select(name_table))
            }

            if field == EVENT:
                self.event_dict = name_dict
            elif field == SITE:
                self.site_dict = name_dict
            elif field == PLAYER:
                self.player_dict = name_dict
            elif field == ANNOTATOR:
                self.annotator_dict = name_dict
            elif field == SOURCE:
                self.source_dict = name_dict

        maxid = self.conn.execute(
            select(func.max(name_table.c.id).label("maxid"))
        ).scalar()
        if maxid is None:
            next_id = 1
        else:
            next_id = maxid + 1

        return next_id

    def do_cancel(self):
        GLib.idle_add(self.progressbar.set_text, "")
        self.cancel = True

    # @profile_me
    def do_import(self, filename, info=None, progressbar=None):
        self.progressbar = progressbar

        orig_filename = filename
        count_source = self.conn.execute(
            select(func.count())
            .select_from(source)
            .where(source.c.name == orig_filename)
        ).scalar()
        if count_source > 0:
            log.info("%s is already imported" % filename)
            return

        # collect new names not in they dict yet
        self.event_data = []
        self.site_data = []
        self.player_data = []
        self.annotator_data = []
        self.source_data = []

        # collect new games and commit them in big chunks for speed
        self.game_data = []
        self.tag_game_data = []

        if filename.startswith("http"):
            filename = download_file(filename, progressbar=progressbar)
            if filename is None:
                return
        else:
            if not os.path.isfile(filename):
                log.info("Can't open %s" % filename)
                return

        if filename.lower().endswith(".zip") and zipfile.is_zipfile(filename):
            with zipfile.ZipFile(filename, "r") as zf:
                path = os.path.dirname(filename)
                files = [
                    os.path.join(path, f)
                    for f in zf.namelist()
                    if f.lower().endswith(".pgn")
                ]
                zf.extractall(path)
        else:
            files = [filename]

        for pgnfile in files:
            base_offset = self.chessfile.size if self.append_pgn else 0

            basename = os.path.basename(pgnfile)
            if progressbar is not None:
                GLib.idle_add(progressbar.set_text, _("Reading %s ..." % basename))
            else:
                log.info("Reading %s ..." % pgnfile)

            size = os.path.getsize(pgnfile)
            handle = protoopen(pgnfile)

            # estimated game count
            all_games = max(size / 840, 1)

            get_id = self.get_id

            try:
                i = 0
                for tags in read_games(handle):
                    if not tags:
                        log.info("Empty game #%s" % (i + 1))
                        continue

                    if self.cancel:
                        self.conn.rollback()
                        return

                    fenstr = tags["FEN"]

                    variant = tags["Variant"]
                    if variant:
                        if "fischer" in variant.lower() or "960" in variant:
                            variant = "Fischerandom"
                        else:
                            variant = variant.lower().capitalize()

                    # Fixes for some non statndard Chess960 .pgn
                    if fenstr and variant == "Fischerandom":
                        parts = fenstr.split()
                        parts[0] = parts[0].replace(".", "/").replace("0", "")
                        if len(parts) == 1:
                            parts.append("w")
                            parts.append("-")
                            parts.append("-")
                        fenstr = " ".join(parts)

                    if variant:
                        if variant not in name2variant:
                            log.info("Unknown variant: %s" % variant)
                            continue
                        variant = name2variant[variant].variant
                        if variant == NORMALCHESS:
                            # lichess uses tag [Variant "Standard"]
                            variant = 0
                    else:
                        variant = 0

                    if basename == "eco.pgn":
                        white = tags["Opening"]
                        black = tags["Variation"]
                    else:
                        white = tags["White"]
                        black = tags["Black"]

                    event_id = get_id(tags["Event"], event, EVENT)

                    site_id = get_id(tags["Site"], site, SITE)

                    date = tags["Date"]

                    game_round = tags["Round"]

                    white_id = get_id(white, player, PLAYER)
                    black_id = get_id(black, player, PLAYER)

                    result = tags["Result"]
                    if result in pgn2Const:
                        result = pgn2Const[result]
                    else:
                        result = RUNNING

                    white_elo = tags["WhiteElo"]
                    black_elo = tags["BlackElo"]

                    time_control = tags["TimeControl"]

                    eco = tags["ECO"][:3]

                    fen = tags["FEN"]

                    board_tag = int(tags["Board"]) if "Board" in tags else 0

                    annotator_id = get_id(tags["Annotator"], annotator, ANNOTATOR)

                    source_id = get_id(orig_filename, source, SOURCE, info=info)

                    ply_count = tags["PlyCount"] if "PlyCount" in tags else 0

                    offset = base_offset + int(tags["offset"])

                    self.game_data.append(
                        {
                            "offset": offset,
                            "offset8": (offset >> 3) << 3,
                            "event_id": event_id,
                            "site_id": site_id,
                            "date": date,
                            "round": game_round,
                            "white_id": white_id,
                            "black_id": black_id,
                            "result": result,
                            "white_elo": white_elo,
                            "black_elo": black_elo,
                            "ply_count": ply_count,
                            "eco": eco,
                            "fen": fen,
                            "variant": variant,
                            "board": board_tag,
                            "time_control": time_control,
                            "annotator_id": annotator_id,
                            "source_id": source_id,
                        }
                    )

                    for tag in tags:
                        if (
                            tag not in dedicated_tags
                            and tag not in other_game_tags
                            and tags[tag]
                        ):
                            self.tag_game_data.append(
                                {
                                    "game_id": self.next_id[GAME],
                                    "tag_name": tag,
                                    "tag_value": tags[tag],
                                }
                            )

                    self.next_id[GAME] += 1
                    i += 1

                    if len(self.game_data) >= self.CHUNK:
                        if self.event_data:
                            self.conn.execute(self.ins_event, self.event_data)
                            self.event_data = []

                        if self.site_data:
                            self.conn.execute(self.ins_site, self.site_data)
                            self.site_data = []

                        if self.player_data:
                            self.conn.execute(self.ins_player, self.player_data)
                            self.player_data = []

                        if self.annotator_data:
                            self.conn.execute(self.ins_annotator, self.annotator_data)
                            self.annotator_data = []

                        if self.source_data:
                            self.conn.execute(self.ins_source, self.source_data)
                            self.source_data = []

                        if self.tag_game_data:
                            self.conn.execute(self.ins_tag_game, self.tag_game_data)
                            self.tag_game_data = []

                        self.conn.execute(self.ins_game, self.game_data)
                        self.game_data = []

                        if progressbar is not None:
                            GLib.idle_add(
                                progressbar.set_fraction, i / float(all_games)
                            )
                            GLib.idle_add(
                                progressbar.set_text,
                                _(
                                    "%(counter)s game headers from %(filename)s imported"
                                    % ({"counter": i, "filename": basename})
                                ),
                            )
                        else:
                            log.info(f"From {pgnfile} imported {i}")

                if self.event_data:
                    self.conn.execute(self.ins_event, self.event_data)
                    self.event_data = []

                if self.site_data:
                    self.conn.execute(self.ins_site, self.site_data)
                    self.site_data = []

                if self.player_data:
                    self.conn.execute(self.ins_player, self.player_data)
                    self.player_data = []

                if self.annotator_data:
                    self.conn.execute(self.ins_annotator, self.annotator_data)
                    self.annotator_data = []

                if self.source_data:
                    self.conn.execute(self.ins_source, self.source_data)
                    self.source_data = []

                if self.tag_game_data:
                    self.conn.execute(self.ins_tag_game, self.tag_game_data)
                    self.tag_game_data = []

                if self.game_data:
                    self.conn.execute(self.ins_game, self.game_data)
                    self.game_data = []

                if progressbar is not None:
                    GLib.idle_add(progressbar.set_fraction, i / float(all_games))
                    GLib.idle_add(
                        progressbar.set_text,
                        _(
                            "%(counter)s game headers from %(filename)s imported"
                            % ({"counter": i, "filename": basename})
                        ),
                    )
                else:
                    log.info(f"From {pgnfile} imported {i}")
                self.conn.commit()

                if self.append_pgn:
                    # reopen database to write
                    self.db_handle.close()
                    with protosave(
                        self.chessfile.path, self.append_pgn
                    ) as self.db_handle:
                        log.info(f"Append from {pgnfile} to {self.chessfile.path}")
                        handle.seek(0)
                        self.db_handle.writelines(handle)
                        handle.close()

                    if self.chessfile.scoutfish is not None:
                        # create new .scout from pgnfile we are importing
                        from pychess.Savers.pgn import scoutfish_path

                        args = [scoutfish_path, "make", pgnfile, "%s" % base_offset]
                        output = subprocess.check_output(
                            args, stderr=subprocess.STDOUT
                        ).decode()

                        # append it to our existing one
                        if output.find("Processing...done") > 0:
                            old_scout = self.chessfile.scoutfish.db
                            new_scout = os.path.splitext(pgnfile)[0] + ".scout"

                            with (
                                open(old_scout, "ab") as file1,
                                open(new_scout, "rb") as file2,
                            ):
                                file1.write(file2.read())

                self.chessfile.handle = protoopen(self.chessfile.path)

            except SQLAlchemyError as e:
                log.error(f"Importing {pgnfile} failed! \n{e}")
                self.conn.rollback()


def read_games(handle):
    """Based on chess.pgn.scan_headers() from Niklas Fiekas python-chess"""

    in_comment = False

    game_headers = None
    game_pos = None

    last_pos = 0
    line = handle.readline()

    # scoutfish creates game offsets at previous game end
    line_end_fix = 2 if line.endswith("\r\n") else 1

    while line:
        # Skip single line comments.
        if line.startswith("%"):
            last_pos += len(line)
            line = handle.readline()
            continue

        # Reading a header tag. Parse it and add it to the current headers.
        if not in_comment and line.startswith("["):
            tag_match = TAG_REGEX.match(line)
            if tag_match:
                if game_pos is None:
                    game_headers = collections.defaultdict(str)
                    game_pos = last_pos

                tag_value = tag_match.group(2)
                tag_value = tag_value.replace('\\"', '"')
                tag_value = tag_value.replace("\\\\", "\\")

                if handle.pgn_encoding != PGN_ENCODING:
                    tag_value = tag_value.encode(PGN_ENCODING).decode(
                        handle.pgn_encoding
                    )
                game_headers[tag_match.group(1)] = tag_value

                last_pos += len(line)
                line = handle.readline()
                continue

        # Reading movetext. Update parser state in_comment in order to skip
        # comments that look like header tags.
        if (not in_comment and "{" in line) or (in_comment and "}" in line):
            in_comment = line.rfind("{") > line.rfind("}")

        # Reading movetext. If there were headers, previously, those are now
        # complete and can be yielded.
        if game_pos is not None:
            game_headers["offset"] = max(0, game_pos - line_end_fix)
            yield game_headers
            game_pos = None

        last_pos += len(line)
        line = handle.readline()

    # Yield the headers of the last game.
    if game_pos is not None:
        game_headers["offset"] = max(0, game_pos - line_end_fix)
        yield game_headers



import asyncio
import datetime

from gi.repository import GLib, GObject

from pychess.System.Log import log
from pychess.Utils.const import ADJOURNED, WHITE, BLACK, UNSUPPORTED

from pychess.ic import (
    RATING_TYPES,
    IC_STATUS_PLAYING,
    IC_STATUS_OFFLINE,
    IC_STATUS_UNKNOWN,
    IC_STATUS_AVAILABLE,
    IC_STATUS_IDLE,
    IC_STATUS_EXAMINING,
    IC_STATUS_NOT_AVAILABLE,
    IC_STATUS_BUSY,
    IC_STATUS_RUNNING_SIMUL_MATCH,
    IC_STATUS_IN_TOURNAMENT,
    TITLE_TYPE_DISPLAY_TEXTS,
    TITLE_TYPE_DISPLAY_TEXTS_SHORT,
    TYPE_BLITZ,
    TYPE_STANDARD,
    TYPE_ATOMIC,
    TYPE_LIGHTNING,
    TYPE_BUGHOUSE,
    TYPE_CRAZYHOUSE,
    TYPE_LOSERS,
    TYPE_SUICIDE,
    TYPE_WILD,
    GAME_TYPES_BY_RATING_TYPE,
    TYPE_UNREGISTERED,
    TYPE_COMPUTER,
    TYPE_ADMINISTRATOR,
    DEVIATION_NONE,
    DEVIATION_ESTIMATED,
    DEVIATION_PROVISIONAL,
    GAME_TYPES_BY_FICS_NAME,
    GAME_TYPES,
    TYPE_GRAND_MASTER,
    TYPE_INTERNATIONAL_MASTER,
    TYPE_FIDE_MASTER,
    TYPE_WOMAN_GRAND_MASTER,
    TYPE_WOMAN_INTERNATIONAL_MASTER,
    TYPE_WOMAN_FIDE_MASTER,
)


class AlreadyExistException(Exception):
    pass


def make_sensitive_if_available(button, player):
    if player.isAvailableForGame():
        button.set_property("sensitive", True)
        button.set_property("tooltip-text", "")
    else:
        button.set_property("sensitive", False)
        button.set_property(
            "tooltip-text",
            _("%(player)s is %(status)s")
            % {"player": player.name, "status": player.display_status.lower()},
        )


def make_sensitive_if_playing(button, player):
    status = player.display_status.lower()
    if player.status == IC_STATUS_PLAYING:
        button.set_property("sensitive", True)
    else:
        button.set_property("sensitive", False)
        if player.status != IC_STATUS_OFFLINE:
            status = _("not playing")
    button.set_property(
        "tooltip-text",
        _("%(player)s is %(status)s") % {"player": player.name, "status": status},
    )


def get_player_tooltip_text(player, show_status=True):
    text = "%s" % player.name
    text += "%s" % player.display_titles(long=True)
    if player.blitz:
        text += "\n{}: {}".format(_("Blitz"), player.blitz)
    if player.standard:
        text += "\n{}: {}".format(_("Standard"), player.standard)
    if player.lightning:
        text += "\n{}: {}".format(_("Lightning"), player.lightning)
    if player.atomic:
        text += "\n{}: {}".format(_("Atomic"), player.atomic)
    if player.bughouse:
        text += "\n{}: {}".format(_("Bughouse"), player.bughouse)
    if player.crazyhouse:
        text += "\n{}: {}".format(_("Crazyhouse"), player.crazyhouse)
    if player.losers:
        text += "\n{}: {}".format(_("Losers"), player.losers)
    if player.suicide:
        text += "\n{}: {}".format(_("Suicide"), player.suicide)
    if player.wild:
        text += "\n{}: {}".format(_("Wild"), player.wild)
    if show_status:
        text += "\n%s" % player.display_status
    return text


def player_id(name):
    """Two players are equal if the first 11 characters of their name match.
    This is to facilitate matching players from output of commands like the 'games'
    command which only return the first 11 characters of a player's name"""
    return name[0:11].lower()


class FICSPlayer(GObject.GObject):
    __gsignals__ = {
        "ratings_changed": (GObject.SignalFlags.RUN_FIRST, None, (int, object))
    }

    def __init__(
        self, name, online=False, status=IC_STATUS_UNKNOWN, game=None, titles=None
    ):
        assert isinstance(name, str), name
        assert isinstance(online, bool), online
        GObject.GObject.__init__(self)
        self.player_id = player_id(name)
        self.name = name
        self.online = online
        self._status = status
        self.status = status
        self.game = None
        self.adjournment = False
        self.ratings = [0 for ratingtype in RATING_TYPES]
        self.deviations = [DEVIATION_NONE for ratingtype in RATING_TYPES]
        if titles is None:
            self.titles = set()
        else:
            self.titles = titles

    def long_name(self, game_type=None):
        name = self.name

        if game_type:
            rating = self.getRatingByGameType(game_type)
        else:
            rating = self.getRatingForCurrentGame()
        if rating:
            name += " (%d)" % rating

        title = self.display_titles()
        if title:
            name += "%s" % title
        return name

    def get_online(self):
        return self._online

    def set_online(self, online):
        self._online = online

    online = GObject.property(get_online, set_online)

    @property
    def display_online(self):
        if self.online:
            return _("Online")
        else:
            return _("Offline")

    def get_status(self):
        return self._status

    def set_status(self, status):
        self._previous_status = self._status
        self._status = status

    status = GObject.property(get_status, set_status)

    def restore_previous_status(self):
        self.status = self._previous_status

    @property
    def display_status(self):
        if self.status == IC_STATUS_AVAILABLE:
            return _("Available")
        elif self.status == IC_STATUS_PLAYING:
            status = _("Playing")
            game = self.game
            if game is not None:
                status += " " + game.display_text
            return status
        elif self.status == IC_STATUS_IDLE:
            return _("Idle")
        elif self.status == IC_STATUS_OFFLINE:
            return _("Offline")
        elif self.status == IC_STATUS_EXAMINING:
            return _("Examining")
        elif self.status in (IC_STATUS_NOT_AVAILABLE, IC_STATUS_BUSY):
            return _("Not Available")
        elif self.status == IC_STATUS_RUNNING_SIMUL_MATCH:
            return _("Running Simul Match")
        elif self.status == IC_STATUS_IN_TOURNAMENT:
            return _("In Tournament")
        else:
            return ""

    def get_game(self):
        return self._game

    def set_game(self, game):
        if game is not None and not isinstance(game, FICSMatch):
            raise TypeError(type(game))
        self._game = game

    game = GObject.property(get_game, set_game)

    def get_titles(self):
        return self._titles

    def set_titles(self, titles):
        self._titles = titles

    titles = GObject.property(get_titles, set_titles)

    def display_titles(self, long=False):
        title = ""
        for item in self.titles:
            if long:
                title += " (" + TITLE_TYPE_DISPLAY_TEXTS[item] + ")"
            else:
                title += " (" + TITLE_TYPE_DISPLAY_TEXTS_SHORT[item] + ")"
        return title

    @property
    def blitz(self):
        return self.ratings[TYPE_BLITZ]

    @property
    def standard(self):
        return self.ratings[TYPE_STANDARD]

    @property
    def lightning(self):
        return self.ratings[TYPE_LIGHTNING]

    @property
    def atomic(self):
        return self.ratings[TYPE_ATOMIC]

    @property
    def bughouse(self):
        return self.ratings[TYPE_BUGHOUSE]

    @property
    def crazyhouse(self):
        return self.ratings[TYPE_CRAZYHOUSE]

    @property
    def losers(self):
        return self.ratings[TYPE_LOSERS]

    @property
    def suicide(self):
        return self.ratings[TYPE_SUICIDE]

    @property
    def wild(self):
        return self.ratings[TYPE_WILD]

    def __eq__(self, player):
        if isinstance(self, type(player)) and self.player_id == player.player_id:
            return True
        else:
            return False

    def __hash__(self):
        return hash(self.player_id)

    def __repr__(self):
        rep = "name='%s'" % (self.name + self.display_titles())
        rep += ", id=%s" % (id(self))
        rep += ", online=%s" % repr(self.online)
        rep += ", adjournment=%s" % repr(self.adjournment)
        rep += ", status=%i" % self.status
        game = self.game
        if game is not None:
            rep += ", game.gameno=%d" % game.gameno
            rep += ", game.rated=%s" % game.rated
            rep += ", game.private=" + repr(game.private)
        else:
            rep += ", game=None"
        for rating_type in RATING_TYPES:
            if rating_type in self.ratings:
                rep += ", {}={}".format(
                    GAME_TYPES_BY_RATING_TYPE[rating_type].display_text,
                    repr(self.ratings[rating_type]),
                )
        return "<FICSPlayer " + rep + ">"

    def isAvailableForGame(self):
        if self.status in (
            IC_STATUS_PLAYING,
            IC_STATUS_BUSY,
            IC_STATUS_OFFLINE,
            IC_STATUS_RUNNING_SIMUL_MATCH,
            IC_STATUS_NOT_AVAILABLE,
            IC_STATUS_EXAMINING,
            IC_STATUS_IN_TOURNAMENT,
        ):
            return False
        else:
            return True

    def isObservable(self):
        return self.status == IC_STATUS_EXAMINING or (
            self.status == IC_STATUS_PLAYING
            and self.game is not None
            and not self.game.private
            and self.game.supported
        )

    def isGuest(self):
        return TYPE_UNREGISTERED in self.titles

    def isComputer(self):
        return TYPE_COMPUTER in self.titles

    def isAdmin(self):
        return TYPE_ADMINISTRATOR in self.titles

    def isTitled(self):
        return (
            (TYPE_GRAND_MASTER in self.titles)
            or (TYPE_INTERNATIONAL_MASTER in self.titles)
            or (TYPE_FIDE_MASTER in self.titles)
            or (TYPE_WOMAN_GRAND_MASTER in self.titles)
            or (TYPE_WOMAN_INTERNATIONAL_MASTER in self.titles)
            or (TYPE_WOMAN_FIDE_MASTER in self.titles)
        )

    @classmethod
    def getIconByRating(cls, rating, size=16):
        from pychess.Utils.IconLoader import get_pixbuf

        assert isinstance(rating, int), "rating not an int: %s" % str(rating)
        if rating >= 1900:
            return get_pixbuf("glade/16x16/weather-storm.png")
        elif rating >= 1600:
            return get_pixbuf("glade/16x16/weather-showers.png")
        elif rating >= 1300:
            return get_pixbuf("glade/16x16/weather-overcast.png")
        elif rating >= 1000:
            return get_pixbuf("glade/16x16/weather-few-clouds.png")
        else:
            return get_pixbuf("glade/16x16/weather-clear.png")

    def getIcon(self, size=16, gametype=None):
        from pychess.Utils.IconLoader import load_icon

        assert isinstance(size, int), "size not an int: %s" % str(size)
        if self.isGuest():
            return load_icon(size, "stock_people", "system-users")
        elif self.isComputer():
            return load_icon(size, "computer", "stock_notebook")
        elif self.isAdmin():
            return load_icon(size, "security-high", "stock_book_blue")
        else:
            if gametype:
                rating = self.getRatingByGameType(gametype)
            else:
                rating = self.getStrength()
            return self.getIconByRating(rating, size)

    def getMarkup(self, gametype=None, big=True, long_titles=True):
        markup = "<b>%s</b>" % self.name
        if self.isGuest():
            markup += self.display_titles(long=long_titles)
        else:
            if gametype:
                rating = self.getRatingByGameType(gametype)
            else:
                rating = self.getStrength()
            if rating < 1:
                rating = _("Unrated")
            markup += " (%s)" % rating

            if self.display_titles() != "":
                markup += self.display_titles(long=long_titles)

        if big:
            markup = "<big>" + markup + "</big>"

        return markup

    def getRatingMean(self):
        ratingtotal = 0
        numratings = 0
        for ratingtype in RATING_TYPES:
            try:
                if self.ratings[ratingtype] == 0:
                    continue
            except IndexError:
                print(ratingtype, RATING_TYPES, self.ratings)
                continue
            if self.deviations[ratingtype] == DEVIATION_NONE:
                ratingtotal += self.ratings[ratingtype] * 3
                numratings += 3
            elif self.deviations[ratingtype] == DEVIATION_ESTIMATED:
                ratingtotal += self.ratings[ratingtype] * 2
                numratings += 2
            elif self.deviations[ratingtype] == DEVIATION_PROVISIONAL:
                ratingtotal += self.ratings[ratingtype] * 1
                numratings += 1
        return numratings > 0 and ratingtotal // numratings or 0

    # FIXME: this isn't very accurate because of inflated standard ratings
    # and deflated lightning ratings and needs work
    # IDEA: use rank in addition to rating to determine strength
    def getStrength(self):
        if self.deviations[TYPE_BLITZ] == DEVIATION_NONE:
            return self.ratings[TYPE_BLITZ]
        elif self.deviations[TYPE_LIGHTNING] == DEVIATION_NONE:
            return self.ratings[TYPE_LIGHTNING]
        else:
            return self.getRatingMean()

    def getRatingByGameType(self, game_type):
        try:
            return self.ratings[game_type.rating_type]
        except KeyError:
            return 0
        except IndexError:
            return 0

    def getRatingForCurrentGame(self):
        try:
            return self.getRatingByGameType(self.game.game_type)
        except AttributeError:
            return 0


class FICSPlayers(GObject.GObject):
    __gsignals__ = {
        "FICSPlayerEntered": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "FICSPlayerExited": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
    }

    def __init__(self, connection):
        GObject.GObject.__init__(self)
        self.players = {}
        self.players_cids = {}
        self.connection = connection

    def start(self):
        #        self.connection.fm.connect("fingeringFinished", self.onFinger)
        pass

    def __getitem__(self, name):
        if not isinstance(name, str):
            raise TypeError("%s" % name)
        if name in self.players:
            return self.players[name]
        else:
            raise KeyError

    def __setitem__(self, key, value):
        """key must be str and value must be FICSPlayer object"""
        if not isinstance(key, str):
            raise TypeError
        if not isinstance(value, FICSPlayer):
            raise TypeError
        if value.player_id in self.players:
            raise AlreadyExistException

        self.players[value.player_id] = value
        self.players_cids[value.player_id] = value.connect(
            "notify::online", self.online_changed
        )

    def __delitem__(self, name):
        if not isinstance(name, str):
            raise TypeError
        if name in self.players:
            player = self.players[name]
            if name in self.players_cids:
                if player.handler_is_connected(self.players_cids[name]):
                    player.disconnect(self.players_cids[name])
                del self.players_cids[name]
            del self.players[name]

    def __contains__(self, name):
        if not isinstance(name, str):
            raise TypeError
        if name in self.players:
            return True
        else:
            return False

    def keys(self):
        return self.players.keys()

    def items(self):
        return self.players.items()

    def values(self):
        return self.players.values()

    def online_changed(self, player, prop):
        if player.online:
            GLib.idle_add(
                self.emit,
                "FICSPlayerEntered",
                [
                    player,
                ],
                priority=GLib.PRIORITY_LOW,
            )

    # This method is a temporary hack until ChatWindow/ChatManager are
    # converted to use FICSPlayer references rather than player's names
    def get_online_playernames(self):
        names = []
        players = list(self.values())
        for player in players:
            if player.online:
                names.append(player.name)
        return names

    def get(self, name):
        key = player_id(name)
        if key in self:
            player = self[key]
        else:
            player = FICSPlayer(name)
            try:
                self[key] = player
            except AlreadyExistException:
                player = self[key]
        return player

    def player_disconnected(self, name):
        # log.debug("%s" % player,
        #    extra={"task": (self.connection.username, "player_disconnected")})
        key = player_id(name)
        if key in self:
            player = self[key]
            player.online = False
            player.status = IC_STATUS_OFFLINE
            if (
                not player.adjournment
                and player.name not in self.connection.notify_users
            ):
                del self[player.player_id]
            else:
                log.debug(
                    "Not removing %s" % player,
                    extra={"task": (self.connection.username, "player_disconnected")},
                )
            GLib.idle_add(
                self.emit, "FICSPlayerExited", player, priority=GLib.PRIORITY_LOW
            )
        else:
            log.debug(
                "Not removing %s (it's not in FICSPlayers)" % name,
                extra={"task": (self.connection.username, "player_disconnected")},
            )

    #    def onFinger (self, fm, finger):
    #        player = player_id(finger.getName())
    #        if player in self:
    #            self[player].finger = finger
    #            # TODO: merge ratings and titles from finger object into ficsplayer object


class FICSMatch(GObject.GObject):
    def __init__(self, minutes, inc, rated, game_type):
        assert minutes is None or isinstance(minutes, int), type(minutes)
        assert inc is None or isinstance(inc, int), inc
        assert isinstance(rated, bool), rated
        assert (
            game_type is None
            or game_type is GAME_TYPES_BY_FICS_NAME["wild"]
            or game_type in GAME_TYPES.values()
        ), game_type
        GObject.GObject.__init__(self)
        self.minutes = minutes
        self.inc = inc
        self.rated = rated
        self.game_type = game_type

    def __repr__(self):
        text = f"{self.minutes} {self.inc}"
        text += " %s" % ("rated" if self.rated else "unrated")
        text += " %s" % self.game_type.display_text
        return text

    @property
    def display_rated(self):
        if self.rated:
            return _("Rated")
        else:
            return _("Unrated")

    @property
    def display_timecontrol(self):
        tim = ""
        if self.minutes is not None:
            tim = _("%d min") % self.minutes
        if self.inc is not None and self.inc != 0:
            tim += _(" + %d sec") % self.inc
        return tim

    @property
    def sortable_time(self):
        # http://www.freechess.org/Help/HelpFiles/etime.html
        etime = self.minutes + int(round(self.inc * 2.0 / 3.0))
        return etime


def get_soughtmatch_tooltip_text(sought):
    text = "%s" % sought.player.name
    text += "%s" % sought.player.display_titles(long=True)
    if not sought.player.isGuest():
        text += " (%d)" % sought.player_rating
    text += f"\n{sought.display_rated} {sought.game_type.display_text}"
    text += "\n" + sought.display_timecontrol
    if sought.color:
        text += "\n" + _("%(player)s plays %(color)s") % {
            "player": sought.player.name,
            "color": _("white") if sought.color == "white" else _("black"),
        }
    return text


class FICSSoughtMatch(FICSMatch):
    def __init__(self, index, player, minutes, inc, rated, color, game_type):
        assert index is None or isinstance(index, int), index
        assert isinstance(player, FICSPlayer), player
        FICSMatch.__init__(self, minutes, inc, rated, game_type)
        self.index = index
        self.player = player
        self.color = color  # self.player plays color

    def __hash__(self):
        return self.index

    def __eq__(self, sought):
        if isinstance(self, type(sought)) and hash(self) == hash(sought):
            return True
        else:
            return False

    def __ne__(self, sought):
        return not self == sought

    def __repr__(self):
        text = "%s" % self.index
        text += " %s" % self.player.name
        text += " %s" % FICSMatch.__repr__(self)
        return text

    @property
    def player_rating(self):
        """
        This returns self.player's rating for the type of match being sought.
        If self.player doesn't have a rating for the type of match being
        sought, this returns 0. If the match is untimed we use self.player's
        standard time-control rating if they have one.
        """
        game_type = self.game_type
        if game_type == GAME_TYPES["untimed"]:
            game_type = GAME_TYPES["standard"]
        return self.player.getRatingByGameType(game_type)


def get_challenge_tooltip_text(challenge):
    text = get_soughtmatch_tooltip_text(challenge)
    if challenge.adjourned:
        text += "\n" + _("This is a continuation of an adjourned match")
    return text


class FICSChallenge(FICSSoughtMatch):
    def __init__(
        self, index, player, minutes, inc, rated, color, game_type, adjourned=False
    ):
        FICSSoughtMatch.__init__(
            self, index, player, minutes, inc, rated, color, game_type
        )
        self.adjourned = adjourned


class FICSChallenges(GObject.GObject):
    __gsignals__ = {
        "FICSChallengeIssued": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "FICSChallengeRemoved": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
    }

    def __init__(self, connection):
        GObject.GObject.__init__(self)
        self.connection = connection
        self.challenges = {}

    def start(self):
        self.connection.om.connect("onChallengeAdd", self.onChallengeIssued)
        self.connection.om.connect("onChallengeRemove", self.onChallengeRemoved)
        self.connection.bm.connect("playGameCreated", self.onPlayingGame)

    def __getitem__(self, index):
        if not isinstance(index, int):
            raise TypeError
        return self.challenges[index]

    def __setitem__(self, index, challenge):
        if not isinstance(index, int):
            raise TypeError
        if not isinstance(challenge, FICSSoughtMatch):
            raise TypeError
        if index in self:
            log.warning(
                "FICSChallenges: not overwriting challenge %s" % repr(challenge)
            )
            return
        self.challenges[index] = challenge
        self.emit("FICSChallengeIssued", challenge)

    def __delitem__(self, index):
        if not isinstance(index, int):
            raise TypeError
        try:
            challenge = self.challenges[index]
        except KeyError:
            return
        del self.challenges[index]
        self.emit("FICSChallengeRemoved", challenge)

    def __contains__(self, index):
        if not isinstance(index, int):
            raise TypeError
        if index in self.challenges:
            return True
        else:
            return False

    def clear(self):
        challenges = self.challenges.copy()
        for key in challenges:
            del self[key]

    def onChallengeIssued(self, om, challenge):
        self[challenge.index] = challenge

    def onChallengeRemoved(self, om, index):
        del self[index]

    def onPlayingGame(self, bm, game):
        self.clear()


def get_rating_range_display_text(rmin=0, rmax=9999):
    assert isinstance(rmin, type(int())) and rmin >= 0 and rmin <= 9999, rmin
    assert isinstance(rmax, type(int())) and rmax >= 0 and rmax <= 9999, rmax
    if rmin > 0:
        text = "%d" % rmin
        if rmax == 9999:
            text += "↑"
        else:
            text += "-%d" % rmax
    elif rmax != 9999:
        text = "%d↓" % rmax
    else:
        text = None
    return text


def get_seek_tooltip_text(seek):
    text = get_soughtmatch_tooltip_text(seek)
    rrtext = get_rating_range_display_text(seek.rmin, seek.rmax)
    if rrtext:
        text += "\n{}: {}".format(_("Opponent Rating"), rrtext)
    if not seek.automatic:
        text += "\n%s" % _("Manual Accept")
    return text


class FICSSeek(FICSSoughtMatch):
    def __init__(
        self,
        index,
        player,
        minutes,
        inc,
        rated,
        color,
        game_type,
        rmin=0,
        rmax=9999,
        automatic=True,
        formula=False,
    ):
        FICSSoughtMatch.__init__(
            self, index, player, minutes, inc, rated, color, game_type
        )
        self.rmin = rmin  # minimum rating one has to accept this seek
        self.rmax = rmax  # maximum rating one has to accept this seek
        self.automatic = automatic  # if True, auto accept; otherwise, manual accept
        self.formula = formula  # players' formula will be used to screen responses


class FICSSeeks(GObject.GObject):
    __gsignals__ = {
        "FICSSeekCreated": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "FICSSeekRemoved": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
    }

    def __init__(self, connection):
        GObject.GObject.__init__(self)
        self.connection = connection
        self.seeks = {}

    def start(self):
        self.connection.glm.connect("addSeek", self.onAddSeek)
        self.connection.glm.connect("removeSeek", self.onRemoveSeek)
        self.connection.glm.connect("clearSeeks", self.onClearSeeks)
        self.connection.bm.connect("curGameEnded", self.onCurGameEnded)

    def __getitem__(self, index):
        if not isinstance(index, int):
            raise TypeError
        return self.seeks[index]

    def __setitem__(self, index, seek):
        if not isinstance(index, int):
            raise TypeError
        if not isinstance(seek, FICSSoughtMatch):
            raise TypeError
        if index in self:
            log.warning("FICSSeeks: not overwriting seek %s" % repr(seek))
            return
        self.seeks[index] = seek
        self.emit("FICSSeekCreated", seek)

    def __delitem__(self, index):
        if not isinstance(index, int):
            raise TypeError
        try:
            seek = self.seeks[index]
        except KeyError:
            return
        del self.seeks[index]
        self.emit("FICSSeekRemoved", seek)

    def __contains__(self, index):
        if not isinstance(index, int):
            raise TypeError
        if index in self.seeks:
            return True
        else:
            return False

    def clear(self):
        seeks = self.seeks.copy()
        for key in seeks:
            del self[key]

    def onAddSeek(self, glm, seek):
        self[seek.index] = seek

    def onRemoveSeek(self, glm, index):
        del self[index]

    def onClearSeeks(self, glm):
        self.clear()

    def onCurGameEnded(self, bm, game):
        self.connection.glm.refresh_seeks()


class FICSBoard:
    def __init__(self, wms, bms, fen=None, pgn=None):
        assert isinstance(wms, int), wms
        assert isinstance(bms, int), bms
        self.wms = wms
        self.bms = bms
        #        assert fen != None or pgn != None
        self.fen = fen
        self.pgn = pgn

    def __repr__(self):
        rep = f"wms={self.wms}\nbms={self.bms}\npgn={self.pgn}"
        return rep


class FICSGame(FICSMatch):
    def __init__(
        self,
        wplayer,
        bplayer,
        gameno=None,
        game_type=None,
        rated=False,
        minutes=None,
        inc=None,
        result=None,
        reason=None,
        board=None,
        private=False,
        relation=None,
    ):
        assert isinstance(wplayer, FICSPlayer), wplayer
        assert isinstance(bplayer, FICSPlayer), bplayer
        assert gameno is None or isinstance(gameno, int), gameno
        assert result is None or isinstance(result, int), result
        assert reason is None or isinstance(reason, int), reason
        assert board is None or isinstance(board, FICSBoard), board
        assert isinstance(private, bool), private
        FICSMatch.__init__(self, minutes, inc, rated, game_type)
        self.wplayer = wplayer
        self.bplayer = bplayer
        self.gameno = gameno
        self.result = result
        self.reason = reason
        self.board = board
        self.private = private
        self.relation = relation

        # move(style12) message queue feeded by BoardManager, consumed by balck and white ICPlayer
        self.move_queue = asyncio.Queue()

    def __hash__(self):
        return hash(
            ":".join(
                (
                    self.wplayer.name[0:11].lower(),
                    self.bplayer.name[0:11].lower(),
                    str(self.gameno),
                )
            )
        )

    def __eq__(self, game):
        if isinstance(game, FICSGame) and hash(self) == hash(game):
            return True
        else:
            return False

    def __repr__(self):
        rep = "<FICSGame wplayer={}, bplayer={}".format(
            repr(self.wplayer),
            repr(self.bplayer),
        )
        if self.gameno is not None:
            rep += ", gameno=%d" % self.gameno
        rep += ", game_type=%s" % self.game_type
        rep += self.rated and ", rated=True" or ", rated=False"
        if self.minutes is not None:
            rep += ", minutes=%i" % self.minutes
        if self.inc is not None:
            rep += ", inc=%i" % self.inc
        if self.result is not None:
            rep += ", result=%i" % self.result
        if self.reason is not None:
            rep += ", reason=%i" % self.reason
        rep += ", private=%s>" % repr(self.private)
        return rep

    def get_private(self):
        return self._private

    def set_private(self, private):
        self._private = private

    private = GObject.property(get_private, set_private)

    @property
    def display_text(self):
        text = ""
        gametype = self.game_type
        if gametype is not None:
            text += gametype.display_text
            if self.private:
                text += " (" + _("Private") + ")"
        return text

    def update(self, game):
        if self.rated != game.rated:
            self.rated = game.rated
        if self.private != game.private:
            self.private = game.private
        if game.minutes is not None and self.minutes != game.minutes:
            self.minutes = game.minutes
        if game.inc is not None and self.inc != game.inc:
            self.inc = game.inc
        if (
            game.game_type is not None
            and self.game_type != game.game_type
            and not (
                self.game_type is not None
                and game.game_type is GAME_TYPES_BY_FICS_NAME["wild"]
            )
        ):
            self.game_type = game.game_type
        if game.result is not None and self.result != game.result:
            self.result = game.result
        if game.reason is not None and self.reason != game.reason:
            self.reason = game.reason
        if game.board is not None and self.board != game.board:
            self.board = game.board

    @property
    def supported(self):
        if self.game_type is GAME_TYPES_BY_FICS_NAME["wild"]:
            return True
        elif self.game_type is not None and self.game_type.fics_name in GAME_TYPES:
            return not GAME_TYPES[self.game_type.fics_name].variant_type in UNSUPPORTED
        else:
            return False


class FICSAdjournedGame(FICSGame):
    def __init__(
        self,
        wplayer,
        bplayer,
        our_color=None,
        length=None,
        time=None,
        rated=False,
        game_type=None,
        private=False,
        minutes=None,
        inc=None,
        result=ADJOURNED,
        reason=None,
        board=None,
        relation=None,
        gameno=None,
    ):
        assert our_color is None or our_color in (WHITE, BLACK), our_color
        assert length is None or isinstance(length, int), length
        assert time is None or isinstance(time, datetime.datetime), time
        FICSGame.__init__(
            self,
            wplayer,
            bplayer,
            rated=rated,
            private=private,
            game_type=game_type,
            minutes=minutes,
            inc=inc,
            result=result,
            reason=reason,
            board=board,
            relation=relation,
            gameno=gameno,
        )
        self.our_color = our_color
        self.length = length
        self.time = time
        self.wrating = ""
        self.brating = ""

    def __repr__(self):
        rep = FICSGame.__repr__(self)[0:-1]
        rep = rep.replace("<FICSGame", "<FICSAdjournedGame")
        if self.our_color is not None:
            rep += ", our_color=%i" % self.our_color
        if self.length is not None:
            rep += ", length=%i" % self.length
        if self.time is not None:
            rep += ", time=%s" % self.display_time
        return rep + ">"

    @property
    def display_time(self):
        if self.time is not None:
            return self.time.isoformat(" ")[0:16]

    @property
    def opponent(self):
        if self.our_color == WHITE:
            return self.bplayer
        elif self.our_color == BLACK:
            return self.wplayer


class FICSHistoryGame(FICSGame):
    def __init__(
        self,
        wplayer,
        bplayer,
        time=None,
        rated=False,
        game_type=None,
        private=False,
        minutes=None,
        inc=None,
        result=None,
        reason=None,
        board=None,
        relation=None,
        wrating=None,
        brating=None,
        gameno=None,
        history_no=None,
    ):
        assert time is None or isinstance(time, datetime.datetime), time
        FICSGame.__init__(
            self,
            wplayer,
            bplayer,
            rated=rated,
            private=private,
            game_type=game_type,
            minutes=minutes,
            inc=inc,
            result=result,
            reason=reason,
            board=board,
            relation=relation,
            gameno=gameno,
        )
        self.time = time
        self.wrating = wrating
        self.brating = brating
        self.history_no = history_no

    def __hash__(self):
        return hash(
            ":".join(
                (
                    self.wplayer.name[0:11].lower(),
                    self.bplayer.name[0:11].lower(),
                    str(self.history_no),
                    str(self.time),
                )
            )
        )

    @property
    def display_time(self):
        if self.time is not None:
            return self.time.isoformat(" ")[0:16]


class FICSJournalGame(FICSGame):
    def __init__(
        self,
        wplayer,
        bplayer,
        our_color=None,
        time=None,
        rated=False,
        game_type=None,
        private=False,
        minutes=None,
        inc=None,
        result=None,
        reason=None,
        board=None,
        relation=None,
        wrating=None,
        brating=None,
        gameno=None,
        journal_no=None,
    ):
        assert our_color is None or our_color in (WHITE, BLACK), our_color
        assert time is None or isinstance(time, datetime.datetime), time
        FICSGame.__init__(
            self,
            wplayer,
            bplayer,
            rated=rated,
            private=private,
            game_type=game_type,
            minutes=minutes,
            inc=inc,
            result=result,
            reason=reason,
            board=board,
            relation=relation,
            gameno=gameno,
        )
        self.wrating = wrating
        self.brating = brating
        self.journal_no = journal_no

    def __hash__(self):
        return hash(
            ":".join(
                (
                    self.wplayer.name[0:11].lower(),
                    self.bplayer.name[0:11].lower(),
                    str(self.journal_no),
                )
            )
        )

    @property
    def display_time(self):
        return _("Unknown")


class FICSGames(GObject.GObject):
    __gsignals__ = {
        "FICSGameCreated": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "FICSGameEnded": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "FICSAdjournedGameRemoved": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "FICSHistoryGameRemoved": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "FICSJournalGameRemoved": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
    }

    def __init__(self, connection):
        GObject.GObject.__init__(self)
        self.games = {}
        self.games_by_gameno = {}
        self.adjourned_games = {}
        self.history_games = {}
        self.journal_games = {}
        self.connection = connection

    def start(self):
        self.connection.adm.connect("onAdjournmentsList", self.onAdjournmentsList)
        self.connection.adm.connect("onHistoryList", self.onHistoryList)
        self.connection.adm.connect("onJournalList", self.onJournalList)
        self.connection.bm.connect("curGameEnded", self.onCurGameEnded)

    def __getitem__(self, game):
        if not isinstance(game, FICSGame):
            raise TypeError("Not a FICSGame: %s" % repr(game))
        if hash(game) in self.games:
            return self.games[hash(game)]
        else:
            raise KeyError

    def __setitem__(self, key, value):
        """key and value must be the same game"""
        if not isinstance(key, FICSGame):
            raise TypeError
        if not isinstance(value, FICSGame):
            raise TypeError
        if key != value:
            raise Exception(f"Not the same: {repr(key)} {repr(value)}")
        if hash(value) in self.games:
            raise Exception(f"{repr(value)} already exists in {repr(self)}")
        self.games[hash(value)] = value
        self.games_by_gameno[value.gameno] = value
        if isinstance(value, FICSAdjournedGame):
            self.adjourned_games[hash(value)] = value
        elif isinstance(value, FICSHistoryGame):
            self.history_games[hash(value)] = value
        elif isinstance(value, FICSJournalGame):
            self.journal_games[hash(value)] = value

    def __delitem__(self, game):
        if not isinstance(game, FICSGame):
            raise TypeError(repr(game), type(game))
        if game in self:
            del self.games[hash(game)]
        if game.gameno in self.games_by_gameno:
            del self.games_by_gameno[game.gameno]
        if game in self.adjourned_games:
            del self.adjourned_games[hash(game)]
        elif game in self.history_games:
            del self.history_games[hash(game)]
        elif game in self.journal_games:
            del self.journal_games[hash(game)]

    def __contains__(self, game):
        if not isinstance(game, FICSGame):
            raise TypeError
        if hash(game) in self.games:
            return True
        else:
            return False

    def keys(self):
        return self.games.keys()

    def items(self):
        return self.games.items()

    def values(self):
        return self.games.values()

    def get_game_by_gameno(self, gameno):
        if not isinstance(gameno, int):
            raise TypeError
        return self.games_by_gameno[gameno]

    def get(self, game, create=True, emit=True):
        # TODO: lock
        if game in self:
            self[game].update(game)
            game = self[game]
        elif create:
            self[game] = game
            if emit:
                self.emit(
                    "FICSGameCreated",
                    [
                        game,
                    ],
                )
        else:
            raise KeyError
        return game

    def game_ended(self, game):
        if game in self:
            if not game.move_queue.empty():

                async def coro(game):
                    # we have to give a chance to ICPlayer
                    # to process the latest move(style12) message
                    # and remove game from game list panel also
                    await asyncio.sleep(0)
                    game = self[game]
                    self.emit("FICSGameEnded", game)
                    del self[game]

                asyncio.create_task(coro(game))
            else:
                game = self[game]
                self.emit("FICSGameEnded", game)
                del self[game]

    def onAdjournmentsList(self, adm, adjournments):
        for game in self.adjourned_games.values():
            if game not in adjournments:
                del self[game]
                game.opponent.adjournment = False
                self.emit("FICSAdjournedGameRemoved", game)

    def onHistoryList(self, adm, history):
        for game in self.history_games.values():
            if game not in history:
                del self[game]
                self.emit("FICSHistoryGameRemoved", game)

    def onJournalList(self, adm, journal):
        for game in self.journal_games.values():
            if game not in journal:
                del self[game]
                self.emit("FICSJournalGameRemoved", game)

    def onCurGameEnded(self, bm, game):
        for adjourned_game in self.adjourned_games.values():
            for player in (game.wplayer, game.bplayer):
                if player == adjourned_game.opponent:
                    del self[adjourned_game]
                    adjourned_game.opponent.adjournment = False
                    self.emit("FICSAdjournedGameRemoved", adjourned_game)

        for game in self.history_games.values():
            del self[game]
            self.emit("FICSHistoryGameRemoved", game)
        self.connection.adm.queryHistory()



"""Three-check Chess Variant"""

from pychess.Utils.const import THREECHECKCHESS, VARIANTS_OTHER_NONSTANDARD
from pychess.Utils.Board import Board

THREECHECKSTART = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 3+3 0 1"


class ThreeCheckBoard(Board):
    variant = THREECHECKCHESS
    __desc__ = _("Win by giving check 3 times")
    name = _("Three-check")
    cecp_name = "3check"
    need_initial_board = False
    standard_rules = False
    variant_group = VARIANTS_OTHER_NONSTANDARD

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=THREECHECKSTART, lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)


def checkCount(board, color):
    lboard = board.clone()
    if color != board.color and lboard.hist_move:
        lboard.popMove()
    cc = 3 - board.remaining_checks[board.color]
    while lboard.hist_move:
        if lboard.isChecked():
            cc += 1
        lboard.popMove()
        if lboard.hist_move:
            lboard.popMove()
    return cc
